import pytest
from pytest import approx
import pandas as pd

import calliope
from calliope.test.common.util import build_test_model as build_model


class TestNationalScaleExampleModelSenseChecks:
    def test_group_prod_min(self):
        model = calliope.examples.national_scale(
            scenario='cold_fusion_with_production_share'
        )
        model.run()

        df_carrier_prod = (
            model.get_formatted_array('carrier_prod')
                 .loc[dict(carriers='power')].sum('locs').sum('timesteps')
                 .to_pandas()
        )

        prod_share = (
            df_carrier_prod.loc[['cold_fusion', 'csp']].sum() /
            df_carrier_prod.loc[['ccgt', 'cold_fusion', 'csp']].sum()
        )

        assert prod_share == approx(0.85)

    def test_group_cap_max(self):
        model = calliope.examples.national_scale(
            scenario='cold_fusion_with_capacity_share'
        )
        model.run()

        cap_share = (
            model.get_formatted_array('energy_cap').loc[{'techs': ['cold_fusion', 'csp']}].sum() /
            model.get_formatted_array('energy_cap').loc[{'techs': ['ccgt', 'cold_fusion', 'csp']}].sum()
        )

        assert cap_share == approx(0.2)

    def test_systemwide_equals(self):
        model = calliope.examples.national_scale(
            override_dict={
                'techs.ccgt.constraints.energy_cap_max_systemwide': 10000,
                'techs.ac_transmission.constraints.energy_cap_equals_systemwide': 6000
            }
        )
        model.run()
        # Check that setting `_equals` to a finite value leads to forcing
        assert (
            model.get_formatted_array('energy_cap').loc[{'techs': 'ccgt'}].sum() == 10000
        )
        assert (
            model.get_formatted_array('energy_cap').loc[{'techs': 'ac_transmission:region1'}].sum() == 6000
        )

    def test_reserve_margin(self):
        model = calliope.examples.national_scale(
            scenario='reserve_margin'
        )

        model.run()

        # constraint_string = '-Inf : -1.1 * ( carrier_con[region1::demand_power::power,2005-01-05 16:00:00] + carrier_con[region2::demand_power::power,2005-01-05 16:00:00] ) / timestep_resolution[2005-01-05 16:00:00] - energy_cap[region1::ccgt] - energy_cap[region1-3::csp] - energy_cap[region1-2::csp] - energy_cap[region1-1::csp] :   0.0'

        # FIXME: capture Pyomo's print output...
        # assert constraint_string in model._backend_model.reserve_margin_constraint.pprint()

        assert float(model.results.cost.sum()) == approx(282487.35489)


@pytest.mark.filterwarnings("ignore:(?s).*Integer:calliope.exceptions.ModelWarning")
class TestUrbanScaleMILP:
    def test_asynchronous_prod_con(self):

        def _get_prod_con(model, prod_con):
            return (
                model.get_formatted_array('carrier_{}'.format(prod_con))
                     .loc[{'techs': 'heat_pipes:X1', 'carriers': 'heat'}]
                     .to_pandas().dropna(how='all')
            )
        m = calliope.examples.urban_scale(override_dict={'run.zero_threshold': 1e-6})
        m.run()
        _prod = _get_prod_con(m, 'prod')
        _con = _get_prod_con(m, 'con')
        assert any(((_con < 0) & (_prod > 0)).any()) is True

        m_bin = calliope.examples.urban_scale(
            override_dict={'techs.heat_pipes.constraints.force_asynchronous_prod_con': True,
                           'run.solver_options.mipgap': 0.05,
                           'run.zero_threshold': 1e-6}
        )
        m_bin.run()
        _prod = _get_prod_con(m_bin, 'prod')
        _con = _get_prod_con(m_bin, 'con')
        assert any(((_con < 0) & (_prod > 0)).any()) is False


class TestModelSettings:
    def test_feasibility(self):

        def override(feasibility, cap_val):
            override_dict = {
                'locations.0.techs': {'test_supply_elec': {}, 'test_demand_elec': {}},
                'links.0,1.exists': False,
                # pick a time subset where demand is uniformally -10 throughout
                'model.subset_time': ['2005-01-01 06:00:00', '2005-01-01 08:00:00'],
                'run.ensure_feasibility': feasibility, 'run.bigM': 1e3,
                # Allow setting resource and energy_cap_max/equals to force infeasibility
                'techs.test_supply_elec.constraints': {
                    'resource': cap_val, 'energy_eff': 1, 'energy_cap_equals': 15,
                    'force_resource': True
                }
            }

            return override_dict

        # Feasible case, unmet_demand/unused_supply is deleted
        model_10 = build_model(
            override_dict=override(True, 10),
            scenario='investment_costs'
        )
        model_10.run()
        for i in ['unmet_demand', 'unused_supply']:
            assert hasattr(model_10._backend_model, i)
            assert i not in model_10._model_data.data_vars.keys()

        # Infeasible case, unmet_demand is required
        model_5 = build_model(
            override_dict=override(True, 5),
            scenario='investment_costs'
        )
        model_5.run()
        assert hasattr(model_5._backend_model, 'unmet_demand')
        assert hasattr(model_5._backend_model, 'unused_supply')
        assert model_5._model_data['unmet_demand'].sum() == 15
        assert 'unused_supply' not in model_5._model_data.data_vars.keys()

        # Infeasible case, unused_supply is required
        model_15 = build_model(
            override_dict=override(True, 15),
            scenario='investment_costs'
        )
        model_15.run()
        assert hasattr(model_15._backend_model, 'unmet_demand')
        assert hasattr(model_15._backend_model, 'unused_supply')
        assert model_15._model_data['unmet_demand'].sum() == -15
        assert 'unused_supply' not in model_15._model_data.data_vars.keys()

        assert (
            model_15._backend_model.obj.expr() - model_10._backend_model.obj.expr() ==
            approx(1e3 * 15)
        )

        assert (
            model_5._backend_model.obj.expr() - model_10._backend_model.obj.expr() ==
            approx(1e3 * 15)
        )

        # Infeasible cases = non-optimal termination
        # too much supply
        model = build_model(
            override_dict=override(False, 15),
            scenario='investment_costs'
        )
        model.run()
        assert not hasattr(model._backend_model, 'unmet_demand')
        assert not hasattr(model._backend_model, 'unused_supply')
        assert not model._model_data.attrs['termination_condition'] == 'optimal'

        # too little supply
        model = build_model(
            override_dict=override(False, 5),
            scenario='investment_costs'
        )
        model.run()
        assert not model._model_data.attrs['termination_condition'] == 'optimal'


class TestGroupConstraints:

    def test_no_group_constraint(self):
        model = build_model(model_file="group_constraints.yaml")
        model.run()
        expensive_generation = (
            model.get_formatted_array("carrier_prod")
                 .loc[{'techs': 'expensive_supply'}].sum().item()
        )
        assert expensive_generation == 0

    def test_switched_off_group_constraint(self):
        model = build_model(
            model_file="group_constraints.yaml",
            scenario="switching_off_group_constraint"
        )
        model.run()
        expensive_generation = (
            model.get_formatted_array("carrier_prod")
                 .loc[{'techs': 'expensive_supply'}].sum().item()
        )
        assert expensive_generation == 0

    @pytest.mark.xfail(reason="Check not yet implemented.")
    def test_group_constraint_without_technology(self):
        model = build_model(
            model_file='group_constraints.yaml',
            scenario='group_constraint_without_tech'
        )
        with pytest.raises(calliope.exceptions.ModelError):
            model.run()

    def test_group_constraint_with_several_constraints(self):
        model = build_model(
            model_file='group_constraints.yaml',
            scenario='several_group_constraints'
        )
        model.run()
        expensive_generation = (model.get_formatted_array("carrier_prod")
                                     .to_dataframe()
                                     .reset_index()
                                     .groupby("techs")
                                     .carrier_prod
                                     .sum()
                                     .transform(lambda x: x / x.sum())
                                     .loc["expensive_supply"])
        assert round(expensive_generation, 5) >= 0.8


class TestDemandShareGroupConstraints:
    def test_no_demand_share_constraint(self):
        model = build_model(model_file='demand_share.yaml')
        model.run()
        expensive_generation = (model.get_formatted_array("carrier_prod")
                                     .to_dataframe()
                                     .reset_index()
                                     .groupby("techs")
                                     .carrier_prod
                                     .sum()
                                     .loc["expensive_elec_supply"])
        assert expensive_generation == 0

    def test_systemwide_demand_share_max_constraint(self):
        model = build_model(
            model_file='demand_share.yaml',
            scenario='demand_share_max_systemwide'
        )
        model.run()
        cheap_generation = model.get_formatted_array("carrier_prod").loc[{'techs': "cheap_elec_supply", 'carriers': "electricity"}].sum()
        demand = -1 * model.get_formatted_array("carrier_con").sum()
        # assert share in each timestep is 0.6
        assert (cheap_generation / demand).round(5) <= 0.3

    def test_systemwide_demand_share_min_constraint(self):
        model = build_model(
            model_file='demand_share.yaml',
            scenario='demand_share_min_systemwide'
        )
        model.run()
        expensive_generation = model.get_formatted_array("carrier_prod").loc[{'techs': "expensive_elec_supply", 'carriers': "electricity"}].sum()
        demand = -1 * model.get_formatted_array("carrier_con").sum()
        # assert share in each timestep is 0.6
        assert (expensive_generation / demand).round(5) >= 0.6

    def test_systemwide_demand_share_equals_constraint(self):
        model = build_model(
            model_file='demand_share.yaml',
            scenario='demand_share_equals_systemwide'
        )
        model.run()
        expensive_generation = model.get_formatted_array("carrier_prod").loc[{'techs': "expensive_elec_supply", 'carriers': "electricity"}].sum()
        demand = -1 * model.get_formatted_array("carrier_con").sum()
        # assert share in each timestep is 0.6
        assert (expensive_generation / demand).round(5) == 0.6

    def test_location_specific_demand_share_max_constraint(self):
        model = build_model(
            model_file='demand_share.yaml',
            scenario='demand_share_max_location_0'
        )
        model.run()
        generation = model.get_formatted_array("carrier_prod").sum(dim='timesteps').loc[{'carriers': "electricity"}]
        demand0 = -model.get_formatted_array("carrier_con").loc[{'locs': '0'}].sum().item()
        cheap_generation0 = generation.loc[{'locs': "0", 'techs': "cheap_elec_supply"}].item()
        expensive_generation1 = generation.loc[{'locs': "1", 'techs': "expensive_elec_supply"}].item()
        assert round(cheap_generation0 / demand0, 5) <= 0.3
        assert expensive_generation1 == 0

    def test_location_specific_demand_share_max_constraint_two_techs(self):
        model = build_model(
            model_file='demand_share.yaml',
            scenario='demand_share_max_location_0_two_techs'
        )
        model.run()
        generation = model.get_formatted_array("carrier_prod").sum(dim='timesteps').loc[{'carriers': "electricity"}]
        demand0 = -model.get_formatted_array("carrier_con").loc[{'locs': '0'}].sum().item()
        generation0 = generation.loc[{'locs': "0", 'techs': ["cheap_elec_supply", "normal_elec_supply"]}].sum('techs').item()
        assert round(generation0 / demand0, 5) <= 0.4

    def test_location_specific_demand_share_min_constraint(self):
        model = build_model(
            model_file='demand_share.yaml',
            scenario='demand_share_min_location_0'
        )
        model.run()
        generation = model.get_formatted_array("carrier_prod").sum(dim='timesteps').loc[{'carriers': "electricity"}]
        demand0 = -model.get_formatted_array("carrier_con").loc[{'locs': '0'}].sum().item()
        expensive_generation0 = generation.loc[{'locs': "0", 'techs': "expensive_elec_supply"}].item()
        expensive_generation1 = generation.loc[{'locs': "1", 'techs': "expensive_elec_supply"}].item()
        assert round(expensive_generation0 / demand0, 5) >= 0.6
        assert expensive_generation1 == 0

    def test_multiple_group_constraints(self):
        model = build_model(
            model_file='demand_share.yaml',
            scenario='multiple_constraints'
        )
        model.run()
        generation = (model.get_formatted_array("carrier_prod")
                           .sum(dim=('timesteps', 'locs', 'carriers')))
        demand = -model.get_formatted_array("carrier_con").sum().item()
        cheap_generation = generation.loc[{'techs': 'cheap_elec_supply'}].item()
        expensive_generation = generation.loc[{'techs': 'expensive_elec_supply'}].item()

        assert round(expensive_generation / demand, 5) >= 0.6
        assert round(cheap_generation / demand, 5) <= 0.3

    def test_multiple_group_carriers(self):
        model = build_model(
            model_file='demand_share.yaml',
            scenario='multiple_carriers_max'
        )
        model.run()
        generation = (model.get_formatted_array("carrier_prod")
                           .sum(dim=('timesteps', 'locs')))
        demand = (-model.get_formatted_array("carrier_con")
                        .sum(dim=('timesteps', 'locs')))
        cheap_generation_elec = generation.loc[{'techs': 'cheap_elec_supply', 'carriers': 'electricity'}].item()
        demand_elec = demand.loc[{'techs': 'electricity_demand', 'carriers': 'electricity'}].item()
        cheap_generation_heat = generation.loc[{'techs': 'cheap_heat_supply', 'carriers': 'heat'}].item()
        demand_heat = demand.loc[{'techs': 'heat_demand', 'carriers': 'heat'}].item()

        assert round(cheap_generation_elec / demand_elec, 5) <= 0.3
        assert round(cheap_generation_heat / demand_heat, 5) <= 0.5

    def test_multiple_group_carriers_constraints(self):
        model = build_model(
            model_file='demand_share.yaml',
            scenario='multiple_constraints_carriers'
        )
        model.run()
        generation = (model.get_formatted_array("carrier_prod")
                           .sum(dim=('timesteps', 'locs')))
        demand = (-model.get_formatted_array("carrier_con")
                        .sum(dim=('timesteps', 'locs')))
        cheap_generation_elec = generation.loc[{'techs': 'cheap_elec_supply', 'carriers': 'electricity'}].item()
        expensive_generation_elec = generation.loc[{'techs': 'expensive_elec_supply', 'carriers': 'electricity'}].item()
        demand_elec = demand.loc[{'techs': 'electricity_demand', 'carriers': 'electricity'}].item()
        cheap_generation_heat = generation.loc[{'techs': 'cheap_heat_supply', 'carriers': 'heat'}].item()
        expensive_generation_heat = generation.loc[{'techs': 'expensive_heat_supply', 'carriers': 'heat'}].item()
        demand_heat = demand.loc[{'techs': 'heat_demand', 'carriers': 'heat'}].item()

        assert round(cheap_generation_elec / demand_elec, 5) <= 0.3
        assert round(expensive_generation_elec / demand_elec, 5) >= 0.6
        assert round(cheap_generation_heat / demand_heat, 5) <= 0.5
        assert round(expensive_generation_heat / demand_heat, 5) >= 0.4

    def test_different_locations_per_group_constraint(self):
        model = build_model(
            model_file='demand_share.yaml',
            scenario='different_locations_per_group'
        )
        model.run()
        generation = (model.get_formatted_array("carrier_prod")
                           .sum(dim=('timesteps', 'carriers')))
        demand = (-model.get_formatted_array("carrier_con")
                        .sum(dim=('timesteps')))
        cheap_generation_0 = generation.loc[{'techs': 'cheap_elec_supply', 'locs': '0'}].item()
        expensive_generation_0 = generation.loc[{'techs': 'expensive_elec_supply', 'locs': '0'}].item()
        cheap_generation_1 = generation.loc[{'techs': 'cheap_elec_supply', 'locs': '1'}].item()
        expensive_generation_1 = generation.loc[{'techs': 'expensive_elec_supply', 'locs': '1'}].item()
        demand_elec_0 = demand.loc[{'techs': 'electricity_demand', 'carriers': 'electricity', 'locs': '0'}].item()
        demand_elec_1 = demand.loc[{'techs': 'electricity_demand', 'carriers': 'electricity', 'locs': '1'}].item()

        assert round(expensive_generation_0 / demand_elec_0, 5) >= 0.6
        assert expensive_generation_1 / demand_elec_1 == 0
        assert round((cheap_generation_0 + cheap_generation_1) / (demand_elec_0 + demand_elec_1), 5) <= 0.3

    def test_transmission_not_included_in_demand(self):
        model = build_model(
            model_file='demand_share.yaml',
            scenario='transmission_not_included_in_demand'
        )
        model.run()
        assert model.results.termination_condition == "optimal"
        generation = (model.get_formatted_array("carrier_prod")
                           .sum(dim=('timesteps', 'carriers')))
        demand = (-model.get_formatted_array("carrier_con")
                        .sum(dim=('timesteps', 'carriers')))

        assert (generation.sel(locs="1", techs=["normal_elec_supply", "cheap_elec_supply", "expensive_elec_supply"])
                          .sum(dim="techs")
                          .item()) == pytest.approx(0)

        cheap_elec_supply_0 = generation.sel(locs="0", techs="cheap_elec_supply").item()
        demand_0 = demand.sel(locs="0", techs="electricity_demand").item()

        assert round(cheap_elec_supply_0 / demand_0, 5) <= 0.4

    def test_demand_share_per_timestep_max(self):
        model = build_model(
            model_file='demand_share.yaml',
            scenario='demand_share_per_timestep_max'
        )
        model.run()
        cheap_generation = model.get_formatted_array("carrier_prod").loc[{'techs': "cheap_elec_supply", 'carriers': "electricity"}].sum('locs')
        demand = -1 * model.get_formatted_array("carrier_con").sum('locs')
        # assert share in each timestep is 0.6
        assert ((cheap_generation / demand).round(5) <= 0.3).all()

    def test_demand_share_per_timestep_min(self):
        model = build_model(
            model_file='demand_share.yaml',
            scenario='demand_share_per_timestep_min'
        )
        model.run()
        expensive_generation = model.get_formatted_array("carrier_prod").loc[{'techs': "expensive_elec_supply", 'carriers': "electricity"}].sum('locs')
        demand = -1 * model.get_formatted_array("carrier_con").sum('locs')
        # assert share in each timestep is 0.6
        assert ((expensive_generation / demand).round(5) >= 0.6).all()

    def test_demand_share_per_timestep_equals(self):
        model = build_model(
            model_file='demand_share.yaml',
            scenario='demand_share_per_timestep_equals'
        )
        model.run()
        expensive_generation = model.get_formatted_array("carrier_prod").loc[{'techs': "expensive_elec_supply", 'carriers': "electricity"}].sum('locs')
        demand = -1 * model.get_formatted_array("carrier_con").sum('locs')
        # assert share in each timestep is 0.6
        assert ((expensive_generation / demand).round(5) == 0.6).all()

    def test_demand_share_per_timestep_decision_inf(self):
        model = build_model(
            model_file='demand_share_decision.yaml',
            scenario='demand_share_per_timestep_decision_inf'
        )
        model.run()
        demand = -1 * model.get_formatted_array("carrier_con").loc[{'carriers': "electricity"}].sum('locs').sum('techs').to_pandas()
        supply = model.get_formatted_array("carrier_prod").loc[{'carriers': "electricity"}].sum('locs').to_pandas().T
        shares = supply.div(demand, axis=0)

        assert (shares['cheap_elec_supply'] == 0.1875).all()
        assert (shares['normal_elec_supply'] == 0.8125).all()

    def test_demand_share_per_timestep_decision_simple(self):
        model = build_model(
            model_file='demand_share_decision.yaml',
            scenario='demand_share_per_timestep_decision_simple'
        )
        model.run()
        demand = -1 * model.get_formatted_array("carrier_con").loc[{'carriers': "electricity"}].sum('locs').sum('techs').to_pandas()
        supply = model.get_formatted_array("carrier_prod").loc[{'carriers': "electricity"}].sum('locs').to_pandas().T
        shares = supply.div(demand, axis=0)

        assert (shares['cheap_elec_supply'] == 0.1875).all()
        assert (shares['normal_elec_supply'] == 0.8125).all()

    def test_demand_share_per_timestep_decision_not_one(self):
        model = build_model(
            model_file='demand_share_decision.yaml',
            scenario='demand_share_per_timestep_decision_not_one'
        )
        model.run()
        demand = -1 * model.get_formatted_array("carrier_con").loc[{'carriers': "electricity"}].sum('locs').sum('techs').to_pandas()
        supply = model.get_formatted_array("carrier_prod").loc[{'carriers': "electricity"}].sum('locs').to_pandas().T
        shares = supply.div(demand, axis=0)

        assert (shares['cheap_elec_supply'] == 0.1).all()
        assert (shares['normal_elec_supply'] == 0.9).all()
        assert shares['expensive_elec_supply'].mean() + shares['normal_elec_supply'].mean() == 0.9

    def test_demand_share_per_timestep_decision_per_location(self):
        model = build_model(
            model_file='demand_share_decision.yaml',
            scenario='demand_share_per_timestep_decision_per_location'
        )
        model.run()
        demand = -1 * model.get_formatted_array("carrier_con").loc[{'carriers': "electricity"}].sum('techs').to_series()
        supply = model.get_formatted_array("carrier_prod").loc[{'carriers': "electricity"}].to_series()
        shares = supply.div(demand, axis=0)

        assert (shares.loc[pd.IndexSlice['0', :, :]].unstack()['cheap_elec_supply'] == 0.25).all()
        assert (shares.loc[pd.IndexSlice['0', :, :]].unstack()['normal_elec_supply'] == 0.75).all()

        assert (shares.loc[pd.IndexSlice['1', :, :]].unstack()['cheap_elec_supply'] == 0.125).all()
        assert (shares.loc[pd.IndexSlice['1', :, :]].unstack()['normal_elec_supply'] == 0.875).all()

    def test_demand_share_per_timestep_decision_inf_with_transmission(self):
        model = build_model(
            model_file='demand_share_decision.yaml',
            scenario='demand_share_per_timestep_decision_inf,with_electricity_transmission'
        )
        model.run()
        demand = -1 * model.get_formatted_array("carrier_con").loc[{'carriers': "electricity"}].sum('techs').to_series()
        supply = model.get_formatted_array("carrier_prod").loc[{'carriers': "electricity"}].to_series()
        shares = supply.div(demand, axis=0)

        assert all([i == pytest.approx(0.12373737) for i in shares.loc[pd.IndexSlice['0', :, :]].unstack()['cheap_elec_supply'].values])
        assert all([i == pytest.approx(0.87626263) for i in shares.loc[pd.IndexSlice['0', :, :]].unstack()['normal_elec_supply'].values])
        assert (shares.loc[pd.IndexSlice['1', :, :]].unstack()['electricity_transmission:0'] == 1.0).all()

    def test_demand_share_per_timestep_decision_inf_with_heat_cosntrain_electricity(self):
        model = build_model(
            model_file='demand_share_decision.yaml',
            scenario='demand_share_per_timestep_decision_inf,with_electricity_conversion_tech'
        )
        model.run()
        demand = -1 * model.get_formatted_array("carrier_con").loc[{'carriers': "electricity"}].sum(['locs', 'techs']).to_pandas()
        supply = model.get_formatted_array("carrier_prod").loc[{'carriers': "electricity"}].sum('locs').to_pandas().T
        shares = supply.div(demand, axis=0)

        assert all([i == pytest.approx(0.175926) for i in shares['cheap_elec_supply'].values])
        assert all([i == pytest.approx(0.824074) for i in shares['normal_elec_supply'].values])

    def test_demand_share_per_timestep_decision_inf_with_heat_constrain_heat(self):
        model = build_model(
            model_file='demand_share_decision.yaml',
            scenario='demand_share_per_timestep_decision_inf_with_heat,with_electricity_conversion_tech'
        )
        model.run()
        demand = -1 * model.get_formatted_array("carrier_con").loc[{'carriers': "heat"}].sum(['locs', 'techs']).to_pandas()
        supply = model.get_formatted_array("carrier_prod").loc[{'carriers': "heat"}].sum('locs').to_pandas().T
        shares = supply.div(demand, axis=0)

        assert all([i == pytest.approx(0.5) for i in shares['elec_to_heat'].values])
        assert all([i == pytest.approx(0.5) for i in shares['heating'].values])

    def test_demand_share_per_timestep_decision_inf_with_heat_constrain_heat_and_electricity(self):
        model = build_model(
            model_file='demand_share_decision.yaml',
            scenario='demand_share_per_timestep_decision_inf_with_heat,demand_share_per_timestep_decision_not_one,with_electricity_conversion_tech'
        )
        model.run()
        demand_heat = -1 * model.get_formatted_array("carrier_con").loc[{'carriers': "heat"}].sum(['locs', 'techs']).to_pandas()
        supply_heat = model.get_formatted_array("carrier_prod").loc[{'carriers': "heat"}].sum('locs').to_pandas().T
        shares_heat = supply_heat.div(demand_heat, axis=0)

        demand_elec = -1 * (
            model._model_data.carrier_con.loc[{
                'loc_tech_carriers_con': [
                    i for i in model._model_data.loc_tech_carriers_demand.values
                    if 'electricity' in i
                ]
            }].sum(['loc_tech_carriers_con']).to_pandas()
        )
        supply_elec = model.get_formatted_array("carrier_prod").loc[{'carriers': "electricity"}].sum('locs').to_pandas().T
        shares_elec = supply_elec.div(demand_elec, axis=0)

        assert all([i == pytest.approx(0.5) for i in shares_heat['elec_to_heat'].values])
        assert all([i == pytest.approx(0.5) for i in shares_heat['heating'].values])
        assert all([i == pytest.approx(0.9) for i in shares_elec['normal_elec_supply'].values])
        assert all([round(i, 5) >= 0.1 for i in shares_elec['cheap_elec_supply'].values])


class TestResourceAreaGroupConstraints:

    def test_no_energy_cap_share_constraint(self):
        model = build_model(model_file='resource_area.yaml')
        model.run()
        cheap_resource_area = (model.get_formatted_array("resource_area")
                                    .loc[{'techs': "cheap_supply"}].sum()).item()
        assert cheap_resource_area == 40

    def test_systemwide_resource_area_max_constraint(self):
        model = build_model(
            model_file='resource_area.yaml',
            scenario='resource_area_max_systemwide'
        )
        model.run()
        cheap_resource_area = (model.get_formatted_array("resource_area")
                                    .loc[{'techs': "cheap_supply"}].sum()).item()
        assert cheap_resource_area == 20

    def test_systemwide_resource_area_min_constraint(self):
        model = build_model(
            model_file='resource_area.yaml',
            scenario='resource_area_min_systemwide'
        )
        model.run()
        resource_area = model.get_formatted_array("resource_area")
        assert resource_area.loc[{'techs': "cheap_supply"}].sum().item() == 0
        assert resource_area.loc[{'techs': "expensive_supply"}].sum().item() == 20

    def test_location_specific_resource_area_max_constraint(self):
        model = build_model(
            model_file='resource_area.yaml',
            scenario='resource_area_max_location_0'
        )
        model.run()
        resource_area = model.get_formatted_array("resource_area")
        cheap_resource_area0 = resource_area.loc[{'locs': "0", 'techs': 'cheap_supply'}].item()
        cheap_resource_area1 = resource_area.loc[{'locs': "1", 'techs': 'cheap_supply'}].item()
        assert cheap_resource_area0 == 10
        assert cheap_resource_area1 == 20

    def test_location_specific_resource_area_min_constraint(self):
        model = build_model(
            model_file='resource_area.yaml',
            scenario='resource_area_min_location_0'
        )
        model.run()
        resource_area = model.get_formatted_array("resource_area")
        expensive_resource_area0 = resource_area.loc[{'locs': "0", 'techs': "expensive_supply"}].item()
        expensive_resource_area1 = resource_area.loc[{'locs': "1", 'techs': "expensive_supply"}].item()
        assert expensive_resource_area0 == 10
        assert expensive_resource_area1 == 0


class TestCostCapGroupConstraint:

    def test_systemwide_cost_max_constraint(self):
        model = build_model(
            model_file='model_cost_cap.yaml',
            scenario='cheap_cost_max_systemwide'
        )
        model.run()
        cheap_cost = (model.get_formatted_array('cost')
                           .loc[{'costs': 'monetary', 'techs': 'cheap_polluting_supply'}]).sum().item()
        assert round(cheap_cost, 5) <= 30

    def test_systemwide_cost_investment_max_constraint(self):
        model = build_model(
            model_file='model_cost_cap.yaml',
            scenario='cheap_cost_investment_max_systemwide'
        )
        model.run()
        cheap_cost = (model.get_formatted_array('cost')
                           .loc[{'costs': 'monetary', 'techs': 'cheap_polluting_supply'}]).sum().item()
        cheap_cost_investment = (model.get_formatted_array('cost_investment')
                                      .loc[{'costs': 'monetary', 'techs': 'cheap_polluting_supply'}]).sum().item()
        assert cheap_cost > cheap_cost_investment
        assert round(cheap_cost_investment, 5) <= 4

    def test_systemwide_cost_var_max_constraint(self):
        model = build_model(
            model_file='model_cost_cap.yaml',
            scenario='cheap_cost_var_max_systemwide'
        )
        model.run()
        cheap_cost = (model.get_formatted_array('cost')
                           .loc[{'costs': 'monetary', 'techs': 'cheap_polluting_supply'}]).sum().item()
        cheap_cost_var = (model.get_formatted_array('cost_var')
                               .loc[{'costs': 'monetary', 'techs': 'cheap_polluting_supply'}]).sum().item()
        assert cheap_cost > cheap_cost_var
        assert round(cheap_cost_var, 5) <= 200

    def test_systemwide_cost_min_constraint(self):
        model = build_model(
            model_file='model_cost_cap.yaml',
            scenario='expensive_cost_min_systemwide'
        )
        model.run()
        expensive_cost = (model.get_formatted_array('cost')
                               .loc[{'costs': 'monetary', 'techs': 'expensive_clean_supply'}]).sum().item()
        assert round(expensive_cost, 5) >= 600

    def test_systemwide_cost_equals_constraint(self):
        model = build_model(
            model_file='model_cost_cap.yaml',
            scenario='cheap_cost_equals_systemwide'
        )
        model.run()
        cheap_cost = (model.get_formatted_array('cost')
                           .loc[{'costs': 'monetary', 'techs': 'cheap_polluting_supply'}]).sum().item()
        assert cheap_cost == approx(210)

    def test_location_specific_cost_max_constraint(self):
        model = build_model(
            model_file='model_cost_cap.yaml',
            scenario='cheap_cost_max_location_0'
        )
        model.run()
        cheap_cost0 = (model.get_formatted_array('cost')
                            .loc[{'costs': 'monetary',
                                  'techs': 'cheap_polluting_supply',
                                  'locs': '0'}]).sum().item()
        assert round(cheap_cost0, 5) <= 10

    def test_systemwide_emissions_max_constraint(self):
        model = build_model(
            model_file='model_cost_cap.yaml',
            scenario='emissions_max_systemwide'
        )
        model.run()
        emissions = (model.get_formatted_array('cost')
                          .loc[{'costs': 'emissions'}]).sum().item()
        assert round(emissions, 5) <= 400

    def test_location_specific_emissions_max_constraint(self):
        model = build_model(
            model_file='model_cost_cap.yaml',
            scenario='emissions_max_location_0'
        )
        model.run()
        emissions0 = (model.get_formatted_array('cost')
                           .loc[{'costs': 'emissions',
                                 'locs': '0'}]).sum().item()
        assert round(emissions0, 5) <= 200

    def test_systemwide_clean_emissions_max_constraint(self):
        model = build_model(
            model_file='model_cost_cap.yaml',
            scenario='clean_emissions_max_systemwide'
        )
        model.run()
        clean_emissions = (model.get_formatted_array('cost')
                                .loc[{'costs': 'emissions',
                                      'techs': 'expensive_clean_supply'}]).sum().item()
        assert round(clean_emissions, 5) <= 300

    def test_multiple_costs_constraint(self):
        model = build_model(
            model_file='model_cost_cap.yaml',
            scenario='multiple_costs_constraint'
        )
        model.run()
        emissions = (model.get_formatted_array('cost')
                          .loc[{'costs': 'emissions'}]).sum().item()
        expensive_cost = (model.get_formatted_array('cost')
                               .loc[{'costs': 'monetary',
                                     'techs': 'expensive_clean_supply'}]).sum().item()
        assert round(emissions, 5) <= 400
        assert round(expensive_cost, 5) <= 600

    def test_different_locations_per_cost_group_constraint(self):
        model = build_model(
            model_file='model_cost_cap.yaml',
            scenario='different_locations_per_group'
        )
        model.run()
        cheap_cost = (model.get_formatted_array('cost')
                           .loc[{'costs': 'monetary', 'techs': 'cheap_polluting_supply'}]).sum().item()
        cheap_cost0 = (model.get_formatted_array('cost')
                            .loc[{'costs': 'monetary',
                                  'techs': 'cheap_polluting_supply',
                                  'locs': '0'}]).sum().item()
        assert round(cheap_cost, 5) <= 30
        assert round(cheap_cost0, 5) <= 10

    def test_different_techs_per_cost_group_constraint(self):
        model = build_model(
            model_file='model_cost_cap.yaml',
            scenario='different_techs_per_group'
        )
        model.run()
        emissions = (model.get_formatted_array('cost')
                          .loc[{'costs': 'emissions'}]).sum().item()
        clean_emissions = (model.get_formatted_array('cost')
                                .loc[{'costs': 'emissions',
                                      'techs': 'expensive_clean_supply'}]).sum().item()
        assert round(emissions, 5) <= 400
        assert round(clean_emissions, 5) <= 300


class TestSupplyShareGroupConstraints:

    def test_no_supply_share_constraint(self):
        model = build_model(model_file='supply_share.yaml')
        model.run()
        expensive_generation = (model.get_formatted_array("carrier_prod")
                                     .loc[{'techs': "expensive_supply"}].sum()).item()
        assert expensive_generation == 0

    def test_systemwide_supply_share_max_constraint(self):
        model = build_model(
            model_file='supply_share.yaml',
            scenario='supply_share_max_systemwide'
        )
        model.run()
        cheap_generation = (model.get_formatted_array("carrier_prod")
                                 .to_dataframe()
                                 .reset_index()
                                 .groupby("techs")
                                 .carrier_prod
                                 .sum()
                                 .transform(lambda x: x / x.sum())
                                 .loc["cheap_supply"])
        assert round(cheap_generation, 5) <= 0.4

    def test_systemwide_supply_share_min_constraint(self):
        model = build_model(
            model_file='supply_share.yaml',
            scenario='supply_share_min_systemwide'
        )
        model.run()
        expensive_generation = (model.get_formatted_array("carrier_prod")
                                     .to_dataframe()
                                     .reset_index()
                                     .groupby("techs")
                                     .carrier_prod
                                     .sum()
                                     .transform(lambda x: x / x.sum())
                                     .loc["expensive_supply"])
        assert round(expensive_generation, 5) >= 0.6

    def test_location_specific_supply_share_max_constraint(self):
        model = build_model(
            model_file='supply_share.yaml',
            scenario='supply_share_max_location_0'
        )
        model.run()
        generation = (model.get_formatted_array("carrier_prod")
                           .sum(dim='timesteps').loc[{'carriers': 'electricity'}])
        cheap_generation0 = generation.loc[{'locs': "0", 'techs': "cheap_supply"}].item()
        expensive_generation0 = generation.loc[{'locs': "0", 'techs': "expensive_supply"}].item()
        expensive_generation1 = generation.loc[{'locs': "1", 'techs': "expensive_supply"}].item()
        assert round(cheap_generation0 / (cheap_generation0 + expensive_generation0), 5) <= 0.4
        assert expensive_generation1 == 0

    def test_location_specific_supply_share_min_constraint(self):
        model = build_model(
            model_file='supply_share.yaml',
            scenario='supply_share_min_location_0'
        )
        model.run()
        generation = (model.get_formatted_array("carrier_prod")
                           .sum(dim='timesteps').loc[{'carriers': 'electricity'}])
        cheap_generation0 = generation.loc[{'locs': "0", 'techs': "cheap_supply"}].item()
        expensive_generation0 = generation.loc[{'locs': "0", 'techs': "expensive_supply"}].item()
        expensive_generation1 = generation.loc[{'locs': "1", 'techs': "expensive_supply"}].item()
        assert round(expensive_generation0 / (cheap_generation0 + expensive_generation0), 5) >= 0.6
        assert expensive_generation1 == 0

    def test_supply_share_with_transmission(self):
        model = build_model(
            model_file='supply_share.yaml',
            scenario='supply_share_min_systemwide,transmission_link'
        )
        model.run()
        expensive_generation = (
            model.get_formatted_array("carrier_prod")
                 .loc[{"techs": "expensive_supply"}].sum().item()
        )
        supply = (
            model._model_data.carrier_prod
            .loc[{"loc_tech_carriers_prod":
                  model._model_data.loc_tech_carriers_supply_all.values}]
            .sum().item()
        )

        assert round(expensive_generation / supply, 5) >= 0.6

    def test_supply_share_with_storage(self):
        model = build_model(
            model_file='supply_share.yaml',
            scenario='supply_share_min_systemwide,storage_tech'
        )
        model.run()
        expensive_generation = (
            model.get_formatted_array("carrier_prod")
                 .loc[{"techs": "expensive_supply"}].sum().item()
        )
        supply = (
            model._model_data.carrier_prod
            .loc[{"loc_tech_carriers_prod":
                  model._model_data.loc_tech_carriers_supply_all.values}]
            .sum().item()
        )

        assert round(expensive_generation / supply, 5) >= 0.6

    def test_supply_share_per_timestep_max(self):
        model = build_model(
            model_file='supply_share.yaml',
            scenario='supply_share_per_timestep_max'
        )
        model.run()
        cheap_supply = model.get_formatted_array("carrier_prod").loc[{'techs': "cheap_supply", 'carriers': "electricity"}].sum('locs')
        supply = model.get_formatted_array("carrier_prod").loc[{'carriers': "electricity"}].sum('locs').sum('techs')
        # assert share in each timestep is 0.4
        assert ((cheap_supply / supply).round(5) <= 0.4).all()

    def test_supply_share_per_timestep_min(self):
        model = build_model(
            model_file='supply_share.yaml',
            scenario='supply_share_per_timestep_min'
        )
        model.run()
        expensive_supply = model.get_formatted_array("carrier_prod").loc[{'techs': "expensive_supply", 'carriers': "electricity"}].sum('locs')
        supply = model.get_formatted_array("carrier_prod").loc[{'carriers': "electricity"}].sum('locs').sum('techs')
        # assert share in each timestep is 0.6
        assert ((expensive_supply / supply).round(5) >= 0.6).all()

    def test_supply_share_per_timestep_equals(self):
        model = build_model(
            model_file='supply_share.yaml',
            scenario='supply_share_per_timestep_equals'
        )
        model.run()
        expensive_supply = model.get_formatted_array("carrier_prod").loc[{'techs': "expensive_supply", 'carriers': "electricity"}].sum('locs')
        supply = model.get_formatted_array("carrier_prod").loc[{'carriers': "electricity"}].sum('locs').sum('techs')
        # assert share in each timestep is 0.6
        assert ((expensive_supply / supply).round(5) == 0.6).all()


class TestEnergyCapShareGroupConstraints:

    def test_no_energy_cap_share_constraint(self):
        model = build_model(model_file='energy_cap_share.yaml')
        model.run()
        expensive_capacity = (model.get_formatted_array("energy_cap")
                                   .loc[{'techs': "expensive_supply"}].sum())
        assert expensive_capacity == 0

    def test_systemwide_energy_cap_share_max_constraint(self):
        model = build_model(
            model_file='energy_cap_share.yaml',
            scenario='energy_cap_share_max_systemwide'
        )
        model.run()
        cheap_capacity = (model.get_formatted_array("energy_cap")
                               .to_dataframe()
                               .reset_index()
                               .groupby("techs")
                               .energy_cap
                               .sum()
                               .loc[["expensive_supply", "cheap_supply"]]  # remove demand
                               .transform(lambda x: x / x.sum())
                               .loc["cheap_supply"])
        assert cheap_capacity <= 0.4

    def test_systemwide_energy_cap_share_min_constraint(self):
        model = build_model(
            model_file='energy_cap_share.yaml',
            scenario='energy_cap_share_min_systemwide'
        )
        model.run()
        expensive_capacity = (model.get_formatted_array("energy_cap")
                                   .to_dataframe()
                                   .reset_index()
                                   .groupby("techs")
                                   .energy_cap
                                   .sum()
                                   .loc[["expensive_supply", "cheap_supply"]]  # remove demand
                                   .transform(lambda x: x / x.sum())
                                   .loc["expensive_supply"])
        assert expensive_capacity >= 0.6

    def test_location_specific_energy_cap_share_max_constraint(self):
        model = build_model(
            model_file='energy_cap_share.yaml',
            scenario='energy_cap_share_max_location_0'
        )
        model.run()
        capacity = model.get_formatted_array("energy_cap")
        cheap_capacity0 = capacity.loc[{'locs': "0", 'techs': "cheap_supply"}].item()
        expensive_capacity0 = capacity.loc[{'locs': "0", 'techs': "expensive_supply"}].item()
        expensive_capacity1 = capacity.loc[{'locs': "1", 'techs': "expensive_supply"}].item()
        assert cheap_capacity0 / (cheap_capacity0 + expensive_capacity0) <= 0.4
        assert expensive_capacity1 == 0

    def test_location_specific_energy_cap_share_min_constraint(self):
        model = build_model(
            model_file='energy_cap_share.yaml',
            scenario='energy_cap_share_min_location_0'
        )
        model.run()
        capacity = model.get_formatted_array("energy_cap")
        cheap_capacity0 = capacity.loc[{'locs': "0", 'techs': "cheap_supply"}].item()
        expensive_capacity0 = capacity.loc[{'locs': "0", 'techs': "expensive_supply"}].item()
        expensive_capacity1 = capacity.loc[{'locs': "1", 'techs': "expensive_supply"}].item()
        assert expensive_capacity0 / (cheap_capacity0 + expensive_capacity0) >= 0.6
        assert expensive_capacity1 == 0


class TestEnergyCapGroupConstraints:

    def test_no_energy_cap_constraint(self):
        model = build_model(model_file='energy_cap.yaml')
        model.run()
        expensive_capacity = (model.get_formatted_array("energy_cap")
                                   .loc[{'techs': "expensive_supply"}].sum()).item()
        assert expensive_capacity == 0

    def test_systemwide_energy_cap_max_constraint(self):
        model = build_model(
            model_file='energy_cap.yaml',
            scenario='energy_cap_max_systemwide'
        )
        model.run()
        cheap_capacity = (model.get_formatted_array("energy_cap")
                               .loc[{'techs': "cheap_supply"}].sum()).item()
        assert round(cheap_capacity, 5) <= 14

    def test_systemwide_energy_cap_min_constraint(self):
        model = build_model(
            model_file='energy_cap.yaml',
            scenario='energy_cap_min_systemwide'
        )
        model.run()
        expensive_capacity = (model.get_formatted_array("energy_cap")
                                   .loc[{'techs': "expensive_supply"}].sum()).item()
        assert round(expensive_capacity, 5) >= 6

    def test_location_specific_energy_cap_max_constraint(self):
        model = build_model(
            model_file='energy_cap.yaml',
            scenario='energy_cap_max_location_0'
        )
        model.run()
        capacity = model.get_formatted_array("energy_cap")
        cheap_capacity0 = capacity.loc[{'locs': "0", 'techs': "cheap_supply"}].item()
        expensive_capacity1 = capacity.loc[{'locs': "1", 'techs': "expensive_supply"}].item()
        assert round(cheap_capacity0, 5) <= 4
        assert expensive_capacity1 == 0

    def test_location_specific_energy_cap_min_constraint(self):
        model = build_model(
            model_file='energy_cap.yaml',
            scenario='energy_cap_min_location_0'
        )
        model.run()
        capacity = model.get_formatted_array("energy_cap")
        expensive_capacity0 = capacity.loc[{'locs': "0", 'techs': "expensive_supply"}].item()
        expensive_capacity1 = capacity.loc[{'locs': "1", 'techs': "expensive_supply"}].item()
        assert round(expensive_capacity0, 5) >= 6
        assert expensive_capacity1 == 0


class TestEnergyCapacityPerStorageCapacity:

    @pytest.fixture
    def model_file(self):
        return "energy_cap_per_storage_cap.yaml"

    @pytest.mark.filterwarnings("ignore:(?s).*`energy_cap_per_storage_cap_min/max/equals`:calliope.exceptions.ModelWarning")
    def test_no_constraint_set(self, model_file):
        model = build_model(model_file=model_file)
        model.run()
        assert model.results.termination_condition == "optimal"
        energy_capacity = model.get_formatted_array("energy_cap").loc[{'techs': 'my_storage'}].sum().item()
        storage_capacity = model.get_formatted_array("storage_cap").loc[{'techs': 'my_storage'}].sum().item()
        assert energy_capacity == pytest.approx(10)
        assert storage_capacity == pytest.approx(175)
        assert storage_capacity != pytest.approx(1 / 10 * energy_capacity)

    def test_equals(self, model_file):
        model = build_model(model_file=model_file, scenario="equals")
        model.run()
        assert model.results.termination_condition == "optimal"
        energy_capacity = model.get_formatted_array("energy_cap").loc[{'techs': 'my_storage'}].sum().item()
        storage_capacity = model.get_formatted_array("storage_cap").loc[{'techs': 'my_storage'}].sum().item()
        assert storage_capacity == pytest.approx(1 / 10 * energy_capacity)

    def test_max(self, model_file):
        model = build_model(model_file=model_file, scenario="max")
        model.run()
        assert model.results.termination_condition == "optimal"
        energy_capacity = model.get_formatted_array("energy_cap").loc[{'techs': 'my_storage'}].sum().item()
        storage_capacity = model.get_formatted_array("storage_cap").loc[{'techs': 'my_storage'}].sum().item()
        assert energy_capacity == pytest.approx(10)
        assert storage_capacity == pytest.approx(1000)

    def test_min(self, model_file):
        model = build_model(model_file=model_file, scenario="min")
        model.run()
        assert model.results.termination_condition == "optimal"
        energy_capacity = model.get_formatted_array("energy_cap").loc[{'techs': 'my_storage'}].sum().item()
        storage_capacity = model.get_formatted_array("storage_cap").loc[{'techs': 'my_storage'}].sum().item()
        assert energy_capacity == pytest.approx(175)
        assert storage_capacity == pytest.approx(175)

    def test_operate_mode(self, model_file):
        model = build_model(model_file=model_file, scenario="operate_mode_min")
        with pytest.raises(calliope.exceptions.ModelError):
            model.run()



import os
import sys

import pytest

import calliope
from calliope import AttrDict


constraint_sets = AttrDict.from_yaml(os.path.join(os.path.dirname(__file__), 'constraint_sets.yaml'))

_defaults_files = {
    k: os.path.join(os.path.dirname(calliope.__file__), 'config', k + '.yaml')
    for k in ['model', 'defaults']
}
defaults = AttrDict.from_yaml(_defaults_files['defaults'])
defaults_model = AttrDict.from_yaml(_defaults_files['model'])


python36_or_higher = pytest.mark.skipif(
    sys.version_info < (3, 6),
    reason="Requires ordered dicts from Python >= 3.6"
)


def build_test_model(override_dict=None, scenario=None, model_file='model.yaml'):
    return calliope.Model(
        os.path.join(os.path.dirname(__file__), 'test_model', model_file),
        override_dict=override_dict,
        scenario=scenario
    )


def check_error_or_warning(error_warning, test_string_or_strings):
    if hasattr(error_warning, 'list'):
        output = ','.join(str(error_warning.list[i]) for i in range(len(error_warning.list)))
    else:
        output = str(error_warning.value)

    if isinstance(test_string_or_strings, list):
        result = all(test_string in output for test_string in test_string_or_strings)
    else:
        result = test_string_or_strings in output

    return result



"""
Copyright (C) 2013-2018 Calliope contributors listed in AUTHORS.
Licensed under the Apache 2.0 License (see LICENSE file).

run_checks.py
~~~~~~~~~~~~~

Checks for model consistency and possible errors when preparing run in the backend.

"""
import numpy as np
import pandas as pd
import xarray as xr
from calliope.core.attrdict import AttrDict
from calliope.core.util.observed_dict import UpdateObserverDict


def check_operate_params(model_data):
    """
    if model mode = `operate`, check for clashes in capacity constraints.
    In this mode, all capacity constraints are set to parameters in the backend,
    so can easily lead to model infeasibility if not checked.

    Returns
    -------
    comments : AttrDict
        debug output
    warnings : list
        possible problems that do not prevent the model run
        from continuing
    errors : list
        serious issues that should raise a ModelError

    """
    defaults = UpdateObserverDict(
        initial_yaml_string=model_data.attrs['defaults'],
        name='defaults', observer=model_data)
    run_config = UpdateObserverDict(
        initial_yaml_string=model_data.attrs['run_config'],
        name='run_config', observer=model_data)

    warnings, errors = [], []
    comments = AttrDict()

    def _get_param(loc_tech, var):
        if _is_in(loc_tech, var) and not any((pd.isnull((model_data[var].loc[loc_tech].values, )),)):
            param = model_data[var].loc[loc_tech].values
        else:
            param = defaults[var]
        return param

    def _is_in(loc_tech, set_or_var):
        try:
            model_data[set_or_var].loc[loc_tech]
            return True
        except (KeyError, AttributeError):
            return False

    for loc_tech in model_data.loc_techs.values:
        energy_cap = model_data.energy_cap.loc[loc_tech].item()
        # Must have energy_cap defined for all relevant techs in the model
        if (pd.isnull(energy_cap) or np.isinf(energy_cap)) and not _is_in(loc_tech, 'force_resource'):
            errors.append(
                'Operate mode: User must define a finite energy_cap (via '
                'energy_cap_equals or energy_cap_max) for {}'.format(loc_tech)
            )

        elif _is_in(loc_tech, 'loc_techs_finite_resource'):
            # force resource overrides capacity constraints, so set capacity constraints to infinity
            if _is_in(loc_tech, 'force_resource'):
                if not _is_in(loc_tech, 'loc_techs_store'):
                    energy_cap = model_data.energy_cap.loc[loc_tech] = np.inf
                    warnings.append(
                        'Energy capacity constraint removed from {} as '
                        'force_resource is applied'.format(loc_tech)
                    )
                if _is_in(loc_tech, 'resource_cap'):
                    print(loc_tech, model_data.resource_cap.loc_techs_supply_plus)
                    model_data.resource_cap.loc[loc_tech] = np.inf
                    warnings.append(
                        'Resource capacity constraint removed from {} as '
                        'force_resource is applied'.format(loc_tech)
                    )
            # Cannot have infinite resource area (physically impossible)
            if _is_in(loc_tech, 'loc_techs_area'):
                area = model_data.resource_area.loc[loc_tech].item()
                if pd.isnull(area) or np.isinf(area):
                    errors.append(
                        'Operate mode: User must define a finite resource_area '
                        '(via resource_area_equals or resource_area_max) for {}, '
                        'as a finite available resource is considered'.format(loc_tech)
                    )
            # Cannot have consumed resource being higher than energy_cap, as
            # constraints will clash. Doesn't affect supply_plus techs with a
            # storage buffer prior to carrier production.
            elif not _is_in(loc_tech, 'loc_techs_store'):
                resource_scale = _get_param(loc_tech, 'resource_scale')
                energy_cap_scale = _get_param(loc_tech, 'energy_cap_scale')
                resource_eff = _get_param(loc_tech, 'resource_eff')
                energy_eff = _get_param(loc_tech, 'energy_eff')
                resource = model_data.resource.loc[loc_tech].values
                if (energy_cap is not None and
                    any(resource * resource_scale * resource_eff >
                        energy_cap * energy_cap_scale * energy_eff)):
                    errors.append(
                        'Operate mode: resource is forced to be higher than '
                        'fixed energy cap for `{}`'.format(loc_tech)
                    )
        if _is_in(loc_tech, 'loc_techs_store'):
            if _is_in(loc_tech, 'charge_rate'):
                storage_cap = model_data.storage_cap.loc[loc_tech].item()
                if storage_cap and energy_cap:
                    charge_rate = model_data['charge_rate'].loc[loc_tech].item()
                    if storage_cap * charge_rate < energy_cap:
                        errors.append(
                            'fixed storage capacity * charge rate is not larger '
                            'than fixed energy capacity for loc::tech {}'.format(loc_tech)
                        )
        if _is_in(loc_tech, 'loc_techs_store'):
            if _is_in(loc_tech, 'energy_cap_per_storage_cap_max'):
                storage_cap = model_data.storage_cap.loc[loc_tech].item()
                if storage_cap and energy_cap:
                    energy_cap_per_storage_cap_max = model_data['energy_cap_per_storage_cap_max'].loc[loc_tech].item()
                    if storage_cap * energy_cap_per_storage_cap_max < energy_cap:
                        errors.append(
                            'fixed storage capacity * energy_cap_per_storage_cap_max is not larger '
                            'than fixed energy capacity for loc::tech {}'.format(loc_tech)
                        )
            elif _is_in(loc_tech, 'energy_cap_per_storage_cap_min'):
                storage_cap = model_data.storage_cap.loc[loc_tech].item()
                if storage_cap and energy_cap:
                    energy_cap_per_storage_cap_min = model_data['energy_cap_per_storage_cap_min'].loc[loc_tech].item()
                    if storage_cap * energy_cap_per_storage_cap_min > energy_cap:
                        errors.append(
                            'fixed storage capacity * energy_cap_per_storage_cap_min is not smaller '
                            'than fixed energy capacity for loc::tech {}'.format(loc_tech)
                        )
    # Must define a resource capacity to ensure the Pyomo param is created
    # for it. But we just create an array of infs, so the capacity has no effect
    if ('resource_cap' not in model_data.data_vars.keys() and
            'loc_techs_supply_plus' in model_data.dims.keys()):
        model_data['resource_cap'] = xr.DataArray(
            [np.inf for i in model_data.loc_techs_supply_plus.values],
            dims='loc_techs_supply_plus')
        model_data['resource_cap'].attrs['is_result'] = 1
        model_data['resource_cap'].attrs['operate_param'] = 1
        warnings.append(
            'Resource capacity constraint defined and set to infinity '
            'for all supply_plus techs'
        )

    window = run_config.get('operation', {}).get('window', None)
    horizon = run_config.get('operation', {}).get('horizon', None)
    if not window or not horizon:
        errors.append(
            'Operational mode requires a timestep window and horizon to be '
            'defined under run.operation'
        )
    elif horizon < window:
        errors.append(
            'Iteration horizon must be larger than iteration window, '
            'for operational mode'
        )

    # Cyclic storage isn't really valid in operate mode, so we ignore it, using
    # initial_storage instead (allowing us to pass storage between operation windows)
    if run_config.get('cyclic_storage', True):
        warnings.append(
            'Storage cannot be cyclic in operate run mode, setting '
            '`run.cyclic_storage` to False for this run'
        )
        run_config['cyclic_storage'] = False

    if 'group_demand_share_per_timestep_decision' in model_data.data_vars:
        warnings.append(
            '`demand_share_per_timestep_decision` group constraints cannot be '
            'used in operate mode, so will not be built.'
        )
        del model_data['group_demand_share_per_timestep_decision']

    return comments, warnings, errors



"""
Copyright (C) 2013-2019 Calliope contributors listed in AUTHORS.
Licensed under the Apache 2.0 License (see LICENSE file).

cli.py
~~~~~~

Command-line interface.

"""

import contextlib
import datetime
import itertools
import logging
import os
import pstats
import shutil
import sys
import traceback

import click

from calliope import AttrDict, Model, examples, read_netcdf
from calliope._version import __version__
from calliope.core.util.generate_runs import generate
from calliope.core.util.logging import logger, set_log_level
from calliope.exceptions import BackendError

_time_format = '%Y-%m-%d %H:%M:%S'


_debug = click.option(
    '--debug', is_flag=True, default=False,
    help='Print debug information when encountering errors.'
)

_quiet = click.option(
    '--quiet', is_flag=True, default=False,
    help='Be less verbose about what is happening, including hiding '
         'solver output.'
)

_pdb = click.option(
    '--pdb', is_flag=True, default=False,
    help='If used together with --debug, drop into interactive '
         'debugger on encountering errors.'
)

_profile = click.option(
    '--profile', is_flag=True, default=False,
    help='Run through cProfile.'
)

_profile_filename = click.option(
    '--profile_filename', type=str,
    help='Filename to save profile to if enabled --profile.'
)


_fail_when_infeasible = click.option(
    '--fail_when_infeasible/--no_fail_when_infeasible', is_flag=True, default=True,
    help='Return fail on command line when problem is infeasible (default True).'
)


if logger.hasHandlers():
    for handler in logger.handlers:
        logger.removeHandler(handler)

formatter = logging.Formatter(
    '[%(asctime)s] %(levelname)-8s %(message)s', datefmt=_time_format
)
console = logging.StreamHandler(stream=sys.stderr)
console.setFormatter(formatter)
logger.addHandler(console)


@contextlib.contextmanager
def format_exceptions(
        debug=False, pdb=False, profile=False,
        profile_filename=None, start_time=None):
    if debug:
        set_log_level('DEBUG')

    try:
        if profile:
            import cProfile
            profile = cProfile.Profile()
            profile.enable()
        yield
        if profile:
            profile.disable()
            if profile_filename:
                dump_path = os.path.expanduser(profile_filename)
                click.secho('\nSaving cProfile output to: {}'.format(dump_path))
                profile.dump_stats(dump_path)
            else:
                click.secho('\n\n----PROFILE OUTPUT----\n\n')
                stats = pstats.Stats(profile).sort_stats('cumulative')
                stats.print_stats(20)  # Print first 20 lines

    except Exception as e:
        if debug:
            traceback.print_exc()
            if pdb:
                import ipdb
                ipdb.post_mortem(e.__traceback__)
        else:
            stack = traceback.extract_tb(e.__traceback__)
            # Get last stack trace entry still in Calliope
            last = [i for i in stack if 'calliope' in i[0]][-1]
            if debug:
                err_string = '\nError in {}, {}:{}'.format(last[2], last[0], last[1])
            else:
                err_string = '\nError in {}:'.format(last[2])
            click.secho(err_string, fg='red')
            click.secho(str(e), fg='red')
            if start_time:
                print_end_time(start_time, msg='aborted due to an error')
        sys.exit(1)


def set_quietness_level(quiet):
    if quiet:
        set_log_level('WARNING')
    else:
        set_log_level('SOLVER')


def print_end_time(start_time, msg='complete'):
    end_time = datetime.datetime.now()
    secs = round((end_time - start_time).total_seconds(), 1)
    tend = end_time.strftime(_time_format)
    click.secho(
        '\nCalliope run {}. '
        'Elapsed: {} seconds (time at exit: {})'.format(msg, secs, tend))


def _get_version():
    return 'Version {}'.format(__version__)


def _cli_start(debug, quiet):
    """
    Initial setup for CLI commands.
    Returns ``start_time`` (datetime timestamp)

    """
    if debug:
        click.secho(_get_version())

    set_quietness_level(quiet)

    logging.captureWarnings(True)
    pywarning_logger = logging.getLogger('py.warnings')
    pywarning_logger.addHandler(console)

    start_time = datetime.datetime.now()

    return start_time


@click.group(invoke_without_command=True)
@click.pass_context
@click.option('--version', is_flag=True, default=False,
              help='Display version.')
def cli(ctx, version):
    """Calliope: a multi-scale energy systems modelling framework"""
    if ctx.invoked_subcommand is None and not version:
        click.secho(ctx.get_help())
    if version:
        click.secho(_get_version())


@cli.command(name='new', short_help='Create a new model based on a built-in example.')
@click.argument('path')
@click.option('--template', type=str, default=None)
@_debug
def new(path, template, debug):
    """
    Create new model at the given ``path``, based on one of the built-in
    example models. The target path must not yet exist. Intermediate
    directories will be created automatically.
    """
    _cli_start(debug, quiet=False)

    with format_exceptions(debug):
        if template is None:
            template = 'national_scale'
        source_path = examples._PATHS[template]
        click.echo('Copying {} template to target directory: {}'.format(template, path))
        shutil.copytree(source_path, path)


def _run_setup_model(
        model_file, scenario, model_format, override_dict):
    """
    Build model in CLI commands. Returns ``model``, a ready-to-run
    calliope.Model instance.

    """
    # Try to determine model file type if not given explicitly
    if model_format is None:
        if model_file.split('.')[-1] in ['yaml', 'yml']:
            model_format = 'yaml'
        elif model_file.split('.')[-1] in ['nc', 'nc4', 'netcdf']:
            model_format = 'netcdf'
        else:
            raise ValueError(
                'Cannot determine model file format based on file '
                'extension for "{}". Set format explicitly with '
                '--model_format.'.format(model_file)
            )

    if model_format == 'yaml':
        model = Model(
            model_file, scenario=scenario, override_dict=override_dict
        )
    elif model_format == 'netcdf':
        if scenario is not None or override_dict is not None:
            raise ValueError(
                'When loading a pre-built model from NetCDF, the '
                '--scenario and --override_dict options are not available.'
            )
        model = read_netcdf(model_file)
    else:
        raise ValueError('Invalid model format: {}'.format(model_format))

    return model


@cli.command(name='run', short_help='Build and run a model.')
@click.argument('model_file')
@click.option('--scenario')
@click.option('--model_format')
@click.option('--override_dict')
@click.option('--save_netcdf')
@click.option('--save_csv')
@click.option('--save_plots')
@click.option('--save_logs')
@click.option(
    '--save_lp', help='Build and save model to the given LP file. '
    'When this is set, the model is not sent to a solver, and all other save options are ignored.')
@_debug
@_quiet
@_pdb
@_profile
@_profile_filename
@_fail_when_infeasible
def run(model_file, scenario, model_format, override_dict,
        save_netcdf, save_csv, save_plots, save_logs, save_lp,
        debug, quiet, pdb, profile, profile_filename, fail_when_infeasible):
    """
    Execute the given model. Tries to guess from the file extension whether
    ``model_file`` is a YAML file or a pre-built model saved to NetCDF.
    This can also explicitly be set with the --model_format=yaml or
    --model_format=netcdf option.

    """
    start_time = _cli_start(debug, quiet)
    click.secho(
        'Calliope {} starting at {}\n'.format(
            __version__,
            start_time.strftime(_time_format)
        ))

    with format_exceptions(debug, pdb, profile, profile_filename, start_time):

        model = _run_setup_model(model_file, scenario, model_format, override_dict)
        click.secho(model.info() + '\n')

        # Only save LP file
        if save_lp:  # Only save LP file without solving model
            click.secho('Saving model to LP file...')
            if save_csv is not None or save_netcdf is not None or save_plots is not None:
                click.secho(
                    'WARNING: Model will not be solved - ignoring other save options!',
                    fg='red', bold=True)
            model.to_lp(save_lp)
            print_end_time(start_time)

        # Else run the model, then save outputs
        else:
            click.secho('Starting model run...')

            if save_logs:
                model.run_config['save_logs'] = save_logs

            if save_csv is None and save_netcdf is None:
                click.secho(
                    '\n!!!\nWARNING: No options to save results have been '
                    'specified.\nModel will run without saving results!\n!!!\n',
                    fg='red', bold=True
                )

            model.run()
            termination = model._model_data.attrs.get('termination_condition', 'unknown')

            if save_csv:
                click.secho('Saving CSV results to directory: {}'.format(save_csv))
                model.to_csv(save_csv)
            if save_netcdf:
                click.secho('Saving NetCDF results to file: {}'.format(save_netcdf))
                model.to_netcdf(save_netcdf)
            if save_plots:
                if termination == 'optimal':
                    click.secho('Saving HTML file with plots to: {}'.format(save_plots))
                    model.plot.summary(to_file=save_plots)
                else:
                    click.secho(
                        'Model termination condition non-optimal. Not saving plots',
                        fg='red', bold=True
                    )

            print_end_time(start_time)
            if fail_when_infeasible and termination != 'optimal':
                raise BackendError("Problem is infeasible.")


@cli.command(
    name='generate_runs',
    short_help='Generate a script to run multiple models.'
)
@click.argument('model_file')
@click.argument('out_file')
@click.option('--kind', help='One of: "bash", "bsub", "sbatch", or "windows".')
@click.option('--scenarios')
@click.option('--cluster_threads', default=1)
@click.option('--cluster_mem')
@click.option('--cluster_time')
@click.option(
    '--additional_args', default='',
    help='Any additional arguments to pass directly on to `calliope run`.')
@click.option('--override_dict')
@_debug
@_quiet
@_pdb
def generate_runs(
        model_file, out_file, kind, scenarios,
        additional_args, override_dict,
        cluster_threads, cluster_mem, cluster_time,
        debug, quiet, pdb):

    _cli_start(debug, quiet)

    kwargs = dict(
        model_file=model_file,
        out_file=out_file,
        scenarios=scenarios,
        additional_args=additional_args,
        override_dict=override_dict,
        cluster_mem=cluster_mem,
        cluster_time=cluster_time,
        cluster_threads=cluster_threads,
    )

    with format_exceptions(debug, pdb):
        generate(kind, **kwargs)


@cli.command(
    name='generate_scenarios',
    short_help='Generate scenario definitions from given combinations of overrides.'
)
@click.argument('model_file')
@click.argument('out_file')
@click.argument('overrides', nargs=-1)
@click.option('--scenario_name_prefix')
@_debug
@_quiet
@_pdb
def generate_scenarios(
        model_file, out_file, overrides, scenario_name_prefix,
        debug, quiet, pdb):

    _cli_start(debug, quiet)

    with format_exceptions(debug, pdb):
        combinations = list(itertools.product(
            *[i.split(';') for i in overrides]
        ))

        if not scenario_name_prefix:
            scenario_name_prefix = 'scenario_'

        # len(str(x)) gives us the number of digits in x, for padding
        scenario_string = '{}{:0>' + str(len(str(len(combinations)))) + 'd}'

        scenarios = {'scenarios': {
            scenario_string.format(scenario_name_prefix, i + 1): c
            for i, c in enumerate(combinations)}}

        AttrDict(scenarios).to_yaml(out_file)



import pytest
from pytest import approx

import pyomo.core as po

import calliope
from calliope.test.common.util import build_test_model as build_model


class TestCostMinimisationObjective:
    def test_nationalscale_minimize_emissions(self):
        model = calliope.examples.national_scale(
            scenario='minimize_emissions_costs',
            override_dict={
                'model.subset_time': '2005-01-01'
            }
        )
        model.run()

        assert model.results.energy_cap.to_pandas()['region1-2::csp'] == approx(10000.0)
        assert model.results.energy_cap.to_pandas()['region1::ac_transmission:region2'] == approx(10000.0)

        assert model.results.carrier_prod.sum('timesteps').to_pandas()['region1::ccgt::power'] == approx(66530.36492823533)

        assert float(model.results.cost.loc[{'costs': 'emissions'}].sum()) == approx(13129619.1)

    def test_nationalscale_maximize_utility(self):
        model = calliope.examples.national_scale(
            scenario='maximize_utility_costs',
            override_dict={
                'model.subset_time': '2005-01-01'
            }
        )
        model.run()

        assert model.results.energy_cap.to_pandas()['region1-2::csp'] == approx(10000.0)
        assert model.results.energy_cap.to_pandas()['region1::ac_transmission:region2'] == approx(10000.0)

        assert model.results.carrier_prod.sum('timesteps').to_pandas()['region1::ccgt::power'] == approx(115569.4354)

        assert float(model.results.cost.sum()) > 6.6e7

    @pytest.mark.parametrize("scenario,cost_class,weight", [
        ('monetary_objective', ['monetary'], [1]),
        ('emissions_objective', ['emissions'], [1]),
        ('weighted_objective', ['monetary', 'emissions'], [0.9, 0.1])
    ])
    def test_weighted_objective_results(self, scenario, cost_class, weight):
        model = build_model(
            model_file='weighted_obj_func.yaml',
            scenario=scenario
        )
        model.run()
        assert (
            sum(
                model.results.cost.loc[{'costs': cost_class[i]}].sum().item() * weight[i]
                for i in range(len(cost_class))
            ) == approx(po.value(model._backend_model.obj))
        )



import os
import tempfile

import pytest  # pylint: disable=unused-import

import calliope
from calliope import exceptions


class TestIO:
    @pytest.fixture(scope='module')
    def model(self):
        model = calliope.examples.national_scale()
        return model

    def test_save_netcdf(self, model):
        bool_attrs = [
            k for k, v in model._model_data.attrs.items()
            if isinstance(v, bool)
        ]

        with tempfile.TemporaryDirectory() as tempdir:
            out_path = os.path.join(tempdir, 'model.nc')
            model.to_netcdf(out_path)
            assert os.path.isfile(out_path)

        # Ensure that boolean attrs have not changed
        for k in bool_attrs:
            assert isinstance(model._model_data.attrs[k], bool)

    def test_save_csv_dir_mustnt_exist(self, model):
        with tempfile.TemporaryDirectory() as tempdir:
            out_path = os.path.join(tempdir)
            with pytest.raises(FileExistsError):
                model.to_csv(out_path)

    @pytest.mark.parametrize("file_name", sorted([
        'inputs_{}.csv'.format(i)
        for i in calliope.examples.national_scale().inputs.data_vars.keys()
    ]))
    def test_save_csv(self, model, file_name):
        with tempfile.TemporaryDirectory() as tempdir:
            out_path = os.path.join(tempdir, 'out_dir')
            model.to_csv(out_path)
            assert os.path.isfile(os.path.join(out_path, file_name))

            with open(os.path.join(out_path, 'inputs_energy_cap_max_systemwide.csv'), 'r') as f:
                assert 'demand_power' not in f.read()

    def test_save_csv_no_dropna(self, model):
        with tempfile.TemporaryDirectory() as tempdir:
            out_path = os.path.join(tempdir, 'out_dir')
            model.to_csv(out_path, dropna=False)

            with open(os.path.join(out_path, 'inputs_energy_cap_max_systemwide.csv'), 'r') as f:
                assert 'demand_power' in f.read()

    def test_save_csv_not_optimal(self):
        # Not checking for content of warnings here, since
        # check_feasibility-related warnings are tested for in
        # test_example_models
        with pytest.warns(exceptions.ModelWarning):
            model = calliope.examples.national_scale(
                scenario='check_feasibility',
                override_dict={'run.cyclic_storage': False}
            )

        model.run()

        with tempfile.TemporaryDirectory() as tempdir:
            out_path = os.path.join(tempdir, 'out_dir')
            with pytest.warns(exceptions.ModelWarning):
                model.to_csv(out_path, dropna=False)

    def test_solve_save_read_netcdf(self, model):
        model.run()

        with tempfile.TemporaryDirectory() as tempdir:
            out_path = os.path.join(tempdir, 'model.nc')
            model.to_netcdf(out_path)
            assert os.path.isfile(out_path)

            model_from_disk = calliope.read_netcdf(out_path)
            # FIXME test for some data in model_from_disk

    def test_save_lp(self, model):
        with tempfile.TemporaryDirectory() as tempdir:
            out_path = os.path.join(tempdir, 'model.lp')
            model.to_lp(out_path)

            with open(out_path, 'r') as f:
                assert '\nmin \nobj:\n+1 cost(monetary_region1_1__csp)' in f.read()



"""
Copyright (C) 2013-2019 Calliope contributors listed in AUTHORS.
Licensed under the Apache 2.0 License (see LICENSE file).

group.py
~~~~~~~~

Group constraints.

"""

import numpy as np
import pyomo.core as po  # pylint: disable=import-error

from calliope.core.util.logging import logger

ORDER = 20  # order in which to invoke constraints relative to other constraint files


def return_noconstraint(*args):
    logger.debug('group constraint returned NoConstraint: {}'.format(','.join(args)))
    return po.Constraint.NoConstraint


def load_constraints(backend_model):
    model_data_dict = backend_model.__calliope_model_data['data']

    for sense in ['min', 'max', 'equals']:
        if 'group_demand_share_{}'.format(sense) in model_data_dict:
            setattr(
                backend_model, 'group_demand_share_{}_constraint'.format(sense),
                po.Constraint(getattr(backend_model, 'group_names_demand_share_{}'.format(sense)),
                              backend_model.carriers, [sense], rule=demand_share_constraint_rule)
            )

        if 'group_demand_share_per_timestep_{}'.format(sense) in model_data_dict:
            setattr(
                backend_model, 'group_demand_share_per_timestep_{}_constraint'.format(sense),
                po.Constraint(getattr(backend_model, 'group_names_demand_share_per_timestep_{}'.format(sense)),
                              backend_model.carriers,
                              backend_model.timesteps,
                              [sense], rule=demand_share_per_timestep_constraint_rule)
            )

        if 'group_supply_share_{}'.format(sense) in model_data_dict:
            setattr(
                backend_model, 'group_supply_share_{}_constraint'.format(sense),
                po.Constraint(getattr(backend_model, 'group_names_supply_share_{}'.format(sense)),
                              backend_model.carriers, [sense], rule=supply_share_constraint_rule)
            )

        if 'group_supply_share_per_timestep_{}'.format(sense) in model_data_dict:
            setattr(
                backend_model, 'group_supply_share_per_timestep_{}_constraint'.format(sense),
                po.Constraint(getattr(backend_model, 'group_names_supply_share_per_timestep_{}'.format(sense)),
                              backend_model.carriers,
                              backend_model.timesteps,
                              [sense], rule=supply_share_per_timestep_constraint_rule)
            )

    if 'group_demand_share_per_timestep_decision' in model_data_dict:
        backend_model.group_demand_share_per_timestep_decision_main_constraint = po.Constraint(
            backend_model.group_names_demand_share_per_timestep_decision,
            backend_model.carriers,
            backend_model.techs,
            backend_model.timesteps,
            rule=demand_share_per_timestep_decision_main_constraint_rule
        )
        backend_model.group_demand_share_per_timestep_decision_sum_constraint = po.Constraint(
            backend_model.group_names_demand_share_per_timestep_decision,
            backend_model.carriers,
            rule=demand_share_per_timestep_decision_sum_constraint_rule
        )

    if 'group_energy_cap_share_min' in model_data_dict:
        backend_model.group_energy_cap_share_min_constraint = po.Constraint(
            backend_model.group_names_energy_cap_share_min,
            ['min'], rule=energy_cap_share_constraint_rule
        )
    if 'group_energy_cap_share_max' in model_data_dict:
        backend_model.group_energy_cap_share_max_constraint = po.Constraint(
            backend_model.group_names_energy_cap_share_max,
            ['max'], rule=energy_cap_share_constraint_rule
        )
    if 'group_energy_cap_min' in model_data_dict:
        backend_model.group_energy_cap_min_constraint = po.Constraint(
            backend_model.group_names_energy_cap_min,
            ['min'], rule=energy_cap_constraint_rule
        )
    if 'group_energy_cap_max' in model_data_dict:
        backend_model.group_energy_cap_max_constraint = po.Constraint(
            backend_model.group_names_energy_cap_max,
            ['max'], rule=energy_cap_constraint_rule
        )
    if 'group_resource_area_min' in model_data_dict:
        backend_model.group_resource_area_min_constraint = po.Constraint(
            backend_model.group_names_resource_area_min,
            ['min'], rule=resource_area_constraint_rule
        )
    if 'group_resource_area_max' in model_data_dict:
        backend_model.group_resource_area_max_constraint = po.Constraint(
            backend_model.group_names_resource_area_max,
            ['max'], rule=resource_area_constraint_rule
        )

    for sense in ['min', 'max', 'equals']:
        if 'group_cost_{}'.format(sense) in model_data_dict:
            setattr(
                backend_model, 'group_cost_{}_constraint'.format(sense),
                po.Constraint(getattr(backend_model, 'group_names_cost_{}'.format(sense)),
                              backend_model.costs, [sense], rule=cost_cap_constraint_rule)
            )
        if 'group_cost_var_{}'.format(sense) in model_data_dict:
            setattr(
                backend_model, 'group_cost_var_{}_constraint'.format(sense),
                po.Constraint(getattr(backend_model, 'group_names_cost_var_{}'.format(sense)),
                              backend_model.costs, [sense], rule=cost_var_cap_constraint_rule)
            )
        if 'group_cost_investment_{}'.format(sense) in model_data_dict:
            setattr(
                backend_model, 'group_cost_investment_{}_constraint'.format(sense),
                po.Constraint(getattr(backend_model, 'group_names_cost_investment_{}'.format(sense)),
                              backend_model.costs, [sense], rule=cost_investment_cap_constraint_rule)
            )


def equalizer(lhs, rhs, sign):
    if sign == 'max':
        return lhs <= rhs
    elif sign == 'min':
        return lhs >= rhs
    elif sign == 'equals':
        return lhs == rhs
    else:
        raise ValueError('Invalid sign: {}'.format(sign))


def get_demand_share_lhs_and_rhs_loc_tech_carriers(backend_model, group_name, carrier):
    """
    Returns
    -------
    (lhs_loc_tech_carriers, rhs_loc_tech_carriers):
        lhs are the supply technologies, rhs are the demand technologies
    """
    lhs_loc_techs = getattr(
        backend_model,
        'group_constraint_loc_techs_{}'.format(group_name)
    )
    lhs_locs = set(loc_tech.split('::')[0] for loc_tech in lhs_loc_techs)
    lhs_loc_tech_carriers = [
        i for i in backend_model.loc_tech_carriers_prod
        if i.rsplit('::', 1)[0] in lhs_loc_techs and i.split('::')[-1] == carrier
    ]
    rhs_loc_tech_carriers = [
        i for i in backend_model.loc_tech_carriers_demand
        if i.split('::')[0] in lhs_locs and i.split('::')[-1] == carrier
    ]
    return (lhs_loc_tech_carriers, rhs_loc_tech_carriers)


def demand_share_constraint_rule(backend_model, group_name, carrier, what):
    """
    Enforces shares of demand of a carrier to be met by the given groups
    of technologies at the given locations, on average over the entire
    model period. The share is relative to ``demand`` technologies only.

    .. container:: scrolling-wrapper

        .. math::

            \\sum_{loc::tech::carrier \\in given\\_group, timestep \\in timesteps} carrier_{prod}(loc::tech::carrier, timestep) \\leq
            share \\times \\sum_{loc::tech:carrier \\in loc\\_techs\\_demand \\in given\\_locations, timestep\\in timesteps}
            carrier_{con}(loc::tech::carrier, timestep)

    """
    model_data_dict = backend_model.__calliope_model_data['data']
    share = model_data_dict['group_demand_share_{}'.format(what)].get(
        (carrier, group_name), np.nan
    )

    if np.isnan(share):
        return return_noconstraint('demand_share', group_name)
    else:
        lhs_loc_tech_carriers, rhs_loc_tech_carriers = get_demand_share_lhs_and_rhs_loc_tech_carriers(
            backend_model, group_name, carrier
        )

        lhs = sum(
            backend_model.carrier_prod[loc_tech_carrier, timestep]
            for loc_tech_carrier in lhs_loc_tech_carriers
            for timestep in backend_model.timesteps
        )

        rhs = share * -1 * sum(
            backend_model.carrier_con[loc_tech_carrier, timestep]
            for loc_tech_carrier in rhs_loc_tech_carriers
            for timestep in backend_model.timesteps
        )

        return equalizer(lhs, rhs, what)


def demand_share_per_timestep_constraint_rule(backend_model, group_name, carrier, timestep, what):
    """
    Enforces shares of demand of a carrier to be met by the given groups
    of technologies at the given locations, in each timestep.
    The share is relative to ``demand`` technologies only.

    .. container:: scrolling-wrapper

        .. math::

            \\sum_{loc::tech::carrier \\in given\\_group} carrier_{prod}(loc::tech::carrier, timestep) \\leq
            share \\times \\sum_{loc::tech:carrier \\in loc\\_techs\\_demand \\in given\\_locations}
            carrier_{con}(loc::tech::carrier, timestep) for timestep \\in timesteps

    """
    model_data_dict = backend_model.__calliope_model_data['data']
    share = model_data_dict['group_demand_share_per_timestep_{}'.format(what)].get(
        (carrier, group_name), np.nan
    )

    if np.isnan(share):
        return return_noconstraint('demand_share_per_timestep', group_name)
    else:
        lhs_loc_tech_carriers, rhs_loc_tech_carriers = get_demand_share_lhs_and_rhs_loc_tech_carriers(
            backend_model, group_name, carrier
        )

        lhs = sum(
            backend_model.carrier_prod[loc_tech_carrier, timestep]
            for loc_tech_carrier in lhs_loc_tech_carriers
        )

        rhs = share * -1 * sum(
            backend_model.carrier_con[loc_tech_carrier, timestep]
            for loc_tech_carrier in rhs_loc_tech_carriers
        )

        return equalizer(lhs, rhs, what)


def demand_share_per_timestep_decision_main_constraint_rule(backend_model, group_name, carrier, tech, timestep):
    """
    Allows the model to decide on how a fraction demand for a carrier is met
    by the given groups, which will all have the same share in each timestep.
    The share is relative to the actual demand from ``demand`` technologies only.

    The main constraint enforces that the shares are the same in each timestep.

    .. container:: scrolling-wrapper

        .. math::
            \\sum_{loc::tech::carrier \\in given\\_group} carrier_{prod}(loc::tech::carrier, timestep)

            =

            \\sum_{loc::tech::carrier \\in given\\_group}
            required\\_resource(loc::tech::carrier, timestep)

            \\times \\sum_{loc::tech::carrier \\in given\\_group}
            demand\\_share\\_per\\_timestep\\_decision(loc::tech::carrier)

            \\forall timestep \\in timesteps

            \\forall tech \\in techs

    """
    model_data_dict = backend_model.__calliope_model_data['data']
    share_of_carrier_demand = model_data_dict['group_demand_share_per_timestep_decision'].get(
        (carrier, group_name), np.nan
    )

    if np.isnan(share_of_carrier_demand):
        return return_noconstraint('demand_share_per_timestep_decision_main', group_name)
    else:
        # lhs are the supply technologies, rhs are the demand technologies
        lhs_loc_tech_carriers, rhs_loc_tech_carriers = get_demand_share_lhs_and_rhs_loc_tech_carriers(
            backend_model, group_name, carrier
        )
        # Filter the supply loc_tech_carriers by the current tech
        lhs_loc_tech_carriers = [i for i in lhs_loc_tech_carriers if '::{}::'.format(tech) in i]

        # Only techs that are in the given group are considered
        if len(lhs_loc_tech_carriers) == 0:
            return return_noconstraint('demand_share_per_timestep_decision_main', group_name)

        lhs = sum(
            backend_model.carrier_prod[loc_tech_carrier, timestep]
            for loc_tech_carrier in lhs_loc_tech_carriers
        )

        rhs = -1 * sum(
            backend_model.required_resource[rhs_loc_tech_carrier.rsplit('::', 1)[0], timestep]
            for rhs_loc_tech_carrier in rhs_loc_tech_carriers
        ) * sum(
            backend_model.demand_share_per_timestep_decision[lhs_loc_tech_carrier]
            for lhs_loc_tech_carrier in lhs_loc_tech_carriers
        )

        return equalizer(lhs, rhs, 'equals')


def demand_share_per_timestep_decision_sum_constraint_rule(backend_model, group_name, carrier):
    """
    Allows the model to decide on how a fraction of demand for a carrier is met
    by the given groups, which will all have the same share in each timestep.
    The share is relative to the actual demand from ``demand`` technologies only.

    The sum constraint ensures that all decision shares add up to the share of
    carrier demand specified in the constraint.

    This constraint is only applied if the share of carrier demand has been
    set to a not-None value.

 .. container:: scrolling-wrapper

        .. math::
            share = \\sum_{loc::tech::carrier \\in given\\_group}
            demand\\_share\\_per\\_timestep\\_decision(loc::tech::carrier)


    """
    model_data_dict = backend_model.__calliope_model_data['data']
    share_of_carrier_demand = model_data_dict['group_demand_share_per_timestep_decision'].get(
        (carrier, group_name), np.nan
    )

    # If inf was given that means that we don't limit the total share
    if np.isinf(share_of_carrier_demand) or np.isnan(share_of_carrier_demand):
        return return_noconstraint('demand_share_per_timestep_decision_sum', group_name)
    else:
        lhs_loc_tech_carriers, _ = get_demand_share_lhs_and_rhs_loc_tech_carriers(
            backend_model, group_name, carrier
        )

        return share_of_carrier_demand == sum(
            backend_model.demand_share_per_timestep_decision[loc_tech_carrier]
            for loc_tech_carrier in lhs_loc_tech_carriers
        )


def get_supply_share_lhs_and_rhs_loc_techs(backend_model, group_name):
    lhs_loc_techs = getattr(
        backend_model,
        'group_constraint_loc_techs_{}'.format(group_name)
    )
    lhs_locs = [loc_tech.split('::')[0] for loc_tech in lhs_loc_techs]
    rhs_loc_techs = [
        i for i in backend_model.loc_techs_supply_all
        if i.split('::')[0] in lhs_locs
    ]
    return (lhs_loc_techs, rhs_loc_techs)


def supply_share_constraint_rule(backend_model, constraint_group, carrier, what):
    """
    Enforces shares of carrier_prod for groups of technologies and locations,
    on average over the entire model period. The share is relative to
    ``supply`` and ``supply_plus`` technologies only.

    .. container:: scrolling-wrapper

        .. math::

            \\sum_{loc::tech::carrier \\in given\\_group, timestep \\in timesteps} carrier_{prod}(loc::tech::carrier, timestep) \\leq
            share \\times \\sum_{loc::tech:carrier \\in loc\\_tech\\_carriers\\_supply\\_all \\in given\\_locations, timestep\\in timesteps}
            carrier_{prod}(loc::tech::carrier, timestep)

    """
    model_data_dict = backend_model.__calliope_model_data['data']
    share = model_data_dict['group_supply_share_{}'.format(what)][(carrier, constraint_group)]

    if np.isnan(share):
        return return_noconstraint('supply_share', constraint_group)
    else:
        lhs_loc_techs, rhs_loc_techs = get_supply_share_lhs_and_rhs_loc_techs(
            backend_model,
            constraint_group
        )

        lhs = sum(
            backend_model.carrier_prod[loc_tech + '::' + carrier, timestep]
            for loc_tech in lhs_loc_techs
            for timestep in backend_model.timesteps
        )
        rhs = share * sum(
            backend_model.carrier_prod[loc_tech + '::' + carrier, timestep]
            for loc_tech in rhs_loc_techs
            for timestep in backend_model.timesteps
        )

        return equalizer(lhs, rhs, what)


def supply_share_per_timestep_constraint_rule(backend_model, constraint_group, carrier, timestep, what):
    """
    Enforces shares of carrier_prod for groups of technologies and locations,
    in each timestep. The share is relative to ``supply`` and ``supply_plus``
    technologies only.

    .. container:: scrolling-wrapper

        .. math::

            \\sum_{loc::tech::carrier \\in given\\_group} carrier_{prod}(loc::tech::carrier, timestep) \\leq
            share \\times \\sum_{loc::tech:carrier \\in loc\\_tech\\_carriers\\_supply\\_all \\in given\\_locations}
            carrier_{prod}(loc::tech::carrier, timestep) for timestep \\in timesteps

    """
    model_data_dict = backend_model.__calliope_model_data['data']
    share = model_data_dict['group_supply_share_per_timestep_{}'.format(what)][(carrier, constraint_group)]

    if np.isnan(share):
        return return_noconstraint('supply_share_per_timestep', constraint_group)
    else:
        lhs_loc_techs, rhs_loc_techs = get_supply_share_lhs_and_rhs_loc_techs(
            backend_model,
            constraint_group
        )

        lhs = sum(
            backend_model.carrier_prod[loc_tech + '::' + carrier, timestep]
            for loc_tech in lhs_loc_techs
        )
        rhs = share * sum(
            backend_model.carrier_prod[loc_tech + '::' + carrier, timestep]
            for loc_tech in rhs_loc_techs
        )

        return equalizer(lhs, rhs, what)


def energy_cap_share_constraint_rule(backend_model, constraint_group, what):
    """
    Enforces shares of energy_cap for groups of technologies and locations. The
    share is relative to ``supply`` and ``supply_plus`` technologies only.

    .. container:: scrolling-wrapper

        .. math::

            \\sum_{loc::tech \\in given\\_group} energy_{cap}(loc::tech) \\leq
            share \\times \\sum_{loc::tech \\in loc\\_tech\\_supply\\_all \\in given\\_locations} energy_{cap}(loc::tech)
    """
    model_data_dict = backend_model.__calliope_model_data['data']
    share = model_data_dict['group_energy_cap_share_{}'.format(what)][(constraint_group)]

    if np.isnan(share):
        return return_noconstraint('energy_cap_share', constraint_group)
    else:
        lhs_loc_techs = getattr(
            backend_model,
            'group_constraint_loc_techs_{}'.format(constraint_group)
        )
        lhs_locs = [loc_tech.split('::')[0] for loc_tech in lhs_loc_techs]
        rhs_loc_techs = [
            i for i in backend_model.loc_techs_supply_all
            if i.split('::')[0] in lhs_locs
        ]

        lhs = sum(
            backend_model.energy_cap[loc_tech]
            for loc_tech in lhs_loc_techs
        )
        rhs = share * sum(
            backend_model.energy_cap[loc_tech]
            for loc_tech in rhs_loc_techs
        )

        return equalizer(lhs, rhs, what)


def energy_cap_constraint_rule(backend_model, constraint_group, what):
    """
    Enforce upper and lower bounds for energy_cap of energy_cap
    for groups of technologies and locations.

    .. container:: scrolling-wrapper

        .. math::

            \\sum_{loc::tech \\in given\\_group} energy_{cap}(loc::tech) \\leq energy\\_cap\\_max\\\\

            \\sum_{loc::tech \\in given\\_group} energy_{cap}(loc::tech) \\geq energy\\_cap\\_min

    """
    model_data_dict = backend_model.__calliope_model_data['data']
    threshold = model_data_dict['group_energy_cap_{}'.format(what)][(constraint_group)]

    if np.isnan(threshold):
        return return_noconstraint('energy_cap', constraint_group)
    else:
        lhs_loc_techs = getattr(
            backend_model,
            'group_constraint_loc_techs_{}'.format(constraint_group)
        )
        lhs = sum(
            backend_model.energy_cap[loc_tech]
            for loc_tech in lhs_loc_techs
        )
        rhs = threshold

        return equalizer(lhs, rhs, what)


def cost_cap_constraint_rule(backend_model, group_name, cost, what):
    """
    Limit cost for a specific cost class to a certain value,
    i.e. Ɛ-constrained costs,
    for groups of technologies and locations.

    .. container:: scrolling-wrapper

        .. math::

            \\sum{loc::tech \\in loc\\_techs_{group\\_name}, timestep \\in timesteps}
            \\boldsymbol{cost}(cost, loc::tech, timestep)
            \\begin{cases}
                \\leq cost\\_max(cost)
                \\geq cost\\_min(cost)
                = cost\\_equals(cost)
            \\end{cases}

    """

    loc_techs = [i for i in getattr(
        backend_model,
        'group_constraint_loc_techs_{}'.format(group_name)
    ) if i in backend_model.loc_techs_cost]

    model_data_dict = backend_model.__calliope_model_data['data']
    cost_cap = model_data_dict['group_cost_{}'.format(what)].get(
        (cost, group_name), np.nan
    )

    if np.isnan(cost_cap):
        return return_noconstraint('cost_cap', group_name)

    sum_cost = sum(backend_model.cost[cost, loc_tech] for loc_tech in loc_techs)

    return equalizer(sum_cost, cost_cap, what)


def cost_investment_cap_constraint_rule(backend_model, group_name, cost, what):
    """
    Limit investment costs specific to a cost class to a
    certain value, i.e. Ɛ-constrained costs,
    for groups of technologies and locations.

    .. container:: scrolling-wrapper

        .. math::

            \\sum{loc::tech \\in loc\\_techs_{group\\_name}, timestep \\in timesteps}
            \\boldsymbol{cost\\_{investment}}(cost, loc::tech, timestep)
            \\begin{cases}
                \\leq cost\\_investment\\_max(cost)
                \\geq cost\\_investment\\_min(cost)
                = cost\\_investment\\_equals(cost)
            \\end{cases}

    """

    loc_techs = [i for i in getattr(
        backend_model,
        'group_constraint_loc_techs_{}'.format(group_name)
    ) if i in backend_model.loc_techs_investment_cost]

    model_data_dict = backend_model.__calliope_model_data['data']
    cost_cap = model_data_dict['group_cost_investment_{}'.format(what)].get(
        (cost, group_name), np.nan
    )

    if np.isnan(cost_cap):
        return return_noconstraint('cost_investment_cap', group_name)

    sum_cost = sum(backend_model.cost_investment[cost, loc_tech] for loc_tech in loc_techs)

    return equalizer(sum_cost, cost_cap, what)


def cost_var_cap_constraint_rule(backend_model, group_name, cost, what):
    """
    Limit variable costs specific to a cost class
    to a certain value, i.e. Ɛ-constrained costs,
    for groups of technologies and locations.

    .. container:: scrolling-wrapper

        .. math::

            \\sum{loc::tech \\in loc\\_techs_{group\\_name}, timestep \\in timesteps}
            \\boldsymbol{cost\\_{var}}(cost, loc::tech, timestep)
            \\begin{cases}
                \\leq cost\\_var\\_max(cost)
                \\geq cost\\_var\\_min(cost)
                = cost\\_var\\_equals(cost)
            \\end{cases}

    """

    loc_techs = [i for i in getattr(
        backend_model,
        'group_constraint_loc_techs_{}'.format(group_name)
    ) if i in backend_model.loc_techs_om_cost]

    model_data_dict = backend_model.__calliope_model_data['data']
    cost_cap = model_data_dict['group_cost_var_{}'.format(what)].get(
        (cost, group_name), np.nan
    )

    if np.isnan(cost_cap):
        return return_noconstraint('cost_var_cap', group_name)

    sum_cost = sum(
        backend_model.cost_var[cost, loc_tech, timestep]
        for loc_tech in loc_techs for timestep in backend_model.timesteps
    )

    return equalizer(sum_cost, cost_cap, what)


def resource_area_constraint_rule(backend_model, constraint_group, what):
    """
    Enforce upper and lower bounds of resource_area for groups of
    technologies and locations.

    .. container:: scrolling-wrapper

        .. math::

            \\boldsymbol{resource_{area}}(loc::tech) \\leq group\\_resource\\_area\\_max\\\\

            \\boldsymbol{resource_{area}}(loc::tech) \\geq group\\_resource\\_area\\_min

    """
    model_data_dict = backend_model.__calliope_model_data['data']
    threshold = model_data_dict['group_resource_area_{}'.format(what)][(constraint_group)]

    if np.isnan(threshold):
        return return_noconstraint('resource_area', constraint_group)
    else:
        lhs_loc_techs = getattr(
            backend_model,
            'group_constraint_loc_techs_{}'.format(constraint_group)
        )

        lhs = sum(
            backend_model.resource_area[loc_tech]
            for loc_tech in lhs_loc_techs
        )
        rhs = threshold

        return equalizer(lhs, rhs, what)



from io import StringIO
import os

import pytest
import numpy as np
import tempfile
import ruamel.yaml as ruamel_yaml

from calliope.core.attrdict import AttrDict, _MISSING
from calliope.test.common.util import check_error_or_warning


class TestAttrDict:
    @pytest.fixture
    def regular_dict(self):
        d = {'a': 1,
             'b': 2,
             'c': {'x': 'foo',
                   'y': 'bar',
                   'z': {'I': 1,
                         'II': 2}
                   },
             'd': None
             }
        return d

    setup_string = """
        # a comment
        a: 1
        b: 2
        # a comment about `c`
        c:  # a comment inline with `c`
            x: foo  # a comment on foo

            #
            y: bar  #
            z:
                I: 1
                II: 2
        d:
    """

    @pytest.fixture
    def yaml_file(self):
        return StringIO(self.setup_string)

    @pytest.fixture
    def yaml_string(self):
        return self.setup_string

    @pytest.fixture
    def attr_dict(self, regular_dict):
        d = regular_dict
        return AttrDict(d)

    def test_missing_nonzero(self):
        assert _MISSING is not True
        assert _MISSING is not False
        assert _MISSING is not None
        assert _MISSING.__nonzero__() is False

    def test_init_from_nondict(self):
        with pytest.raises(ValueError) as excinfo:
            d = AttrDict('foo')
        assert check_error_or_warning(
            excinfo, 'Must pass a dict to AttrDict'
        )

    def test_init_from_dict(self, regular_dict):
        d = AttrDict(regular_dict)
        assert d.a == 1

    def test_init_from_dict_with_nested_keys(self):
        d = AttrDict({'foo.bar.baz': 1})
        assert d.foo.bar.baz == 1

    def test_from_yaml_fobj(self, yaml_file):
        d = AttrDict.from_yaml(yaml_file)
        assert d.a == 1
        assert d.c.z.II == 2

    def test_from_yaml_path(self):
        this_path = os.path.dirname(__file__)
        yaml_path = os.path.join(this_path, 'common', 'yaml_file.yaml')
        d = AttrDict.from_yaml(yaml_path)
        assert d.a == 1
        assert d.c.z.II == 2

    def test_from_yaml_string(self, yaml_string):
        d = AttrDict.from_yaml_string(yaml_string)
        assert d.a == 1
        assert d.c.z.II == 2

    def test_from_yaml_string_dot_strings(self):
        yaml_string = 'a.b.c: 1\na.b.foo: 2'
        d = AttrDict.from_yaml_string(yaml_string)
        assert d.a.b.c == 1
        assert d.a.b.foo == 2

    def test_from_yaml_string_dot_strings_duplicate(self):
        yaml_string = 'a.b.c: 1\na.b.c: 2'
        with pytest.warns(ruamel_yaml.constructor.DuplicateKeyFutureWarning):
            AttrDict.from_yaml_string(yaml_string)

    def test_simple_invalid_yaml(self):
        yaml_string = '1 this is not valid yaml'
        with pytest.raises(ValueError) as excinfo:
            AttrDict.from_yaml_string(yaml_string)
        assert check_error_or_warning(
            excinfo, 'Could not parse <yaml string> as YAML'
        )

    def test_parser_error(self):
        with pytest.raises(ruamel_yaml.YAMLError):
            AttrDict.from_yaml_string("""
            foo: bar
            baz: 1
                - foobar
                bar: baz

            """)

    def test_dot_access_first(self, attr_dict):
        d = attr_dict
        assert d.a == 1

    def test_dot_access_second(self, attr_dict):
        d = attr_dict
        assert d.c.x == 'foo'

    def test_dot_access_list(self):
        d = AttrDict.from_yaml_string("a: [{x: 1}, {y: 2}]")
        assert d.a[0].x == 1

    def test_set_key_first(self, attr_dict):
        d = attr_dict
        d.set_key('a', 2)
        assert d.a == 2

    def test_set_key_second(self, attr_dict):
        d = attr_dict
        d.set_key('c.x', 'baz')
        assert d.c.x == 'baz'

    def test_set_key_multiple_inexisting(self, attr_dict):
        d = attr_dict
        d.set_key('c.l.o.h.a', 'foo')
        assert d.c.l.o.h.a == 'foo'

    def test_set_key_nested_on_string(self, attr_dict):
        d = attr_dict
        with pytest.raises(KeyError):
            d.set_key('a.foo', 'bar')

    def test_set_key_nested_on_none(self, attr_dict):
        d = attr_dict
        assert d['d'] is None
        d.set_key('d.foo', 'bar')
        assert d.d.foo == 'bar'

    def test_pass_regular_dict_to_set_key(self, attr_dict):
        # Regular dicts get turned into AttrDicts when using
        # assignment through set_key()
        attr_dict.set_key('c.z.newkey', {'foo': 1, 'doo': 2})
        assert isinstance(attr_dict.get_key('c.z.newkey'), AttrDict)
        assert attr_dict.get_key('c.z.newkey.foo') == 1

    def test_get_subkey_from_nested_non_attrdict(self, attr_dict):
        # Directly assigning a dict means it is not modified
        # but it breaks get_key with nested keys
        attr_dict['c']['z']['newkey'] = {'foo': 1, 'doo': 2}

        with pytest.raises(AttributeError) as excinfo:
            attr_dict.get_key('c.z.newkey.foo')

        assert check_error_or_warning(
            excinfo, "'dict' object has no attribute 'get_key'"
        )

    def test_get_key_first(self, attr_dict):
        d = attr_dict
        assert d.get_key('a') == 1

    def test_get_key_second(self, attr_dict):
        d = attr_dict
        assert d.get_key('c.x') == 'foo'

    def test_get_key_inexistant(self, attr_dict):
        d = attr_dict
        with pytest.raises(KeyError):
            d.get_key('foo')

    def test_get_key_second_inexistant(self, attr_dict):
        d = attr_dict
        with pytest.raises(KeyError):
            d.get_key('foo.bar')

    def test_get_key_default(self, attr_dict):
        d = attr_dict
        assert d.get_key('c.x', default='bar') == 'foo'

    def test_get_key_inexistant_default(self, attr_dict):
        d = attr_dict
        assert d.get_key('foo', default='baz') == 'baz'

    def test_get_key_second_inexistant_default(self, attr_dict):
        d = attr_dict
        assert d.get_key('foo.bar', default='baz') == 'baz'

    def test_get_key_second_nondict_default(self, attr_dict):
        d = attr_dict
        assert d.get_key('c.x.foo', default='baz') == 'baz'

    def test_get_key_inexistant_default_false(self, attr_dict):
        d = attr_dict
        assert d.get_key('foo', default=False) is False

    def test_get_key_second_inexistant_default_false(self, attr_dict):
        d = attr_dict
        assert d.get_key('foo.bar', default=False) is False

    def test_as_dict(self, attr_dict):
        d = attr_dict
        dd = d.as_dict()
        assert dd['a'] == 1
        assert dd['c']['x'] == 'foo'

    def test_as_dict_with_sublists(self):
        d = AttrDict.from_yaml_string("a: [{x: 1}, {y: 2}]")
        dd = d.as_dict()
        assert dd['a'][0]['x'] == 1
        assert isinstance(dd['a'][0], dict)  # Not AttrDict!

    def test_as_dict_flat(self, attr_dict):
        dd = attr_dict.as_dict(flat=True)
        assert dd['c.x'] == 'foo'

    def test_keys_nested_as_list(self, attr_dict):
        d = attr_dict
        dd = d.keys_nested()
        assert dd == ['a', 'b', 'c.x', 'c.y', 'c.z.I', 'c.z.II', 'd']

    def test_keys_nested_as_dict(self, attr_dict):
        d = attr_dict
        dd = d.keys_nested(subkeys_as='dict')
        assert dd == ['a', 'b', {'c': ['x', 'y', {'z': ['I', 'II']}]}, 'd']

    def test_union(self, attr_dict):
        d = attr_dict
        d_new = AttrDict()
        d_new.set_key('c.z.III', 'foo')
        d.union(d_new)
        assert d.c.z.III == 'foo'
        assert d.c.z.I == 1

    def test_union_duplicate_keys(self, attr_dict):
        d = attr_dict
        d_new = AttrDict()
        d_new.set_key('c.z.II', 'foo')
        with pytest.raises(KeyError):
            d.union(d_new)

    def test_union_replacement(self, attr_dict):
        d = attr_dict
        d_new = AttrDict.from_yaml_string("""
            c: {_REPLACE_: foo}
        """)
        d.union(d_new, allow_override=True, allow_replacement=True)
        assert d.c == 'foo'

    def test_union_empty_dicts(self, attr_dict):
        d = attr_dict
        d_new = AttrDict({
            '1': {'foo': {}},
            'baz': {'bar': {}},
        })
        d.union(d_new)
        assert len(d.baz.bar.keys()) == 0

    def test_del_key_single(self, attr_dict):
        attr_dict.del_key('c')
        assert 'c' not in attr_dict

    def test_del_key_nested(self, attr_dict):
        attr_dict.del_key('c.z.I')
        assert 'I' not in attr_dict.c.z

    def test_to_yaml(self, yaml_file):
        d = AttrDict.from_yaml(yaml_file)
        d.set_key('numpy.some_int', np.int32(10))
        d.set_key('numpy.some_float', np.float64(0.5))
        d.a_list = [0, 1, 2]
        with tempfile.TemporaryDirectory() as tempdir:
            out_file = os.path.join(tempdir, 'test.yaml')
            d.to_yaml(out_file)

            with open(out_file, 'r') as f:
                result = f.read()

            assert 'some_int: 10' in result
            assert 'some_float: 0.5' in result
            assert 'a_list:\n- 0\n- 1\n- 2' in result

    def test_to_yaml_string(self, yaml_file):
        d = AttrDict.from_yaml(yaml_file)
        result = d.to_yaml()
        assert 'a: 1' in result

    def test_import_must_be_list(self):
        yaml_string = """
            import: 'somefile.yaml'
        """
        with pytest.raises(ValueError) as excinfo:
            AttrDict.from_yaml_string(yaml_string, resolve_imports=True)
        assert check_error_or_warning(
            excinfo, '`import` must be a list.')

    def test_do_not_resolve_imports(self):
        yaml_string = """
            import: ['somefile.yaml']
        """
        d = AttrDict.from_yaml_string(yaml_string, resolve_imports=False)
        # Should not raise an error about a missing file, as we ask for
        # imports not to be resolved
        assert d['import'] == ['somefile.yaml']

    def test_nested_import(self, yaml_file):
        with tempfile.TemporaryDirectory() as tempdir:
            imported_file = os.path.join(tempdir, 'test_import.yaml')
            imported_yaml = """
                somekey: 1
                anotherkey: 2
            """
            with open(imported_file, 'w') as f:
                f.write(imported_yaml)

            yaml_string = """
                foobar:
                    import:
                        - {}
                foo:
                    bar: 1
                    baz: 2
                    3:
                        4: 5
            """.format(imported_file)

            d = AttrDict.from_yaml_string(
                yaml_string, resolve_imports='foobar'
            )

        assert 'foobar.somekey' in d.keys_nested()
        assert d.get_key('foobar.anotherkey') == 2



"""
Copyright (C) 2013-2018 Calliope contributors listed in AUTHORS.
Licensed under the Apache 2.0 License (see LICENSE file).

"""

import logging

import numpy as np
import pandas as pd
import xarray as xr

from calliope.core.util.tools import memoize
from calliope.core.util.logging import logger
from calliope import exceptions


@memoize
def get_param(backend_model, var, dims):
    """
    Get an input parameter held in a Pyomo object, or held in the defaults
    dictionary if that Pyomo object doesn't exist.

    Parameters
    ----------
    backend_model : Pyomo model instance
    var : str
    dims : single value or tuple

    """
    try:
        return getattr(backend_model, var)[dims]
    except AttributeError:  # i.e. parameter doesn't exist at all
        logger.debug('get_param: var {} and dims {} leading to default lookup'.format(var, dims))
        return backend_model.__calliope_defaults[var]
    except KeyError:  # try removing timestep
        try:
            if len(dims) > 2:
                return getattr(backend_model, var)[dims[:-1]]
            else:
                return getattr(backend_model, var)[dims[0]]
        except KeyError:  # Static default value
            logger.debug('get_param: var {} and dims {} leading to default lookup'.format(var, dims))
            return backend_model.__calliope_defaults[var]


def get_previous_timestep(timesteps, timestep):
    """Get the timestamp for the timestep previous to the input timestep"""
    # order_dict starts numbering at zero, timesteps is one-indexed, so we do not need
    # to subtract 1 to get to previous_step -- it happens "automagically"
    return timesteps[timesteps.order_dict[timestep]]


@memoize
def get_loc_tech_carriers(backend_model, loc_carrier):
    """
    For a given loc_carrier concatenation, get lists of the relevant
    loc_tech_carriers which produce energy (loc_tech_carriers_prod), consume
    energy (loc_tech_carriers_con) and export energy (loc_tech_carriers_export)
    """
    lookup = backend_model.__calliope_model_data['data']['lookup_loc_carriers']
    loc_tech_carriers = split_comma_list(lookup[loc_carrier])

    loc_tech_carriers_prod = [
        i for i in loc_tech_carriers if i in backend_model.loc_tech_carriers_prod
    ]
    loc_tech_carriers_con = [
        i for i in loc_tech_carriers if i in backend_model.loc_tech_carriers_con
    ]

    if hasattr(backend_model, 'loc_tech_carriers_export'):
        loc_tech_carriers_export = [
            i for i in loc_tech_carriers
            if i in backend_model.loc_tech_carriers_export
        ]
    else:
        loc_tech_carriers_export = []

    return (
        loc_tech_carriers_prod,
        loc_tech_carriers_con,
        loc_tech_carriers_export
    )


@memoize
def get_loc_tech(loc_tech_carrier):
    """
    Split the string of a loc_tech_carrier (e.g. `region1::ccgt::power`) to get
    just the loc_tech (e.g. `region1::ccgt`)
    """
    return loc_tech_carrier.rsplit('::', 1)[0]


@memoize
def get_timestep_weight(backend_model):
    """
    Get the total number of years this model considers, by summing all
    timestep resolution with timestep weight (a weight/resolution of 1 = 1 hour)
    and divide it by number of hours in the year. Weight/resolution will almost
    always be 1 per step, unless time clustering/masking/resampling has taken place.
    """
    model_data_dict = backend_model.__calliope_model_data
    time_res = list(model_data_dict['data']['timestep_resolution'].values())
    weights = list(model_data_dict['data']['timestep_weights'].values())
    return sum(np.multiply(time_res, weights)) / 8760


@memoize
def split_comma_list(comma_list):
    """
    Take a comma deliminated string and split it into a list of strings
    """
    return comma_list.split(',')


@memoize
def get_conversion_plus_io(backend_model, tier):
    """
    from a carrier_tier, return the primary tier (of `in`, `out`) and
    corresponding decision variable (`carrier_con` and `carrier_prod`, respectively)
    """
    if 'out' in tier:
        return 'out', backend_model.carrier_prod
    elif 'in' in tier:
        return 'in', backend_model.carrier_con


def get_var(backend_model, var, dims=None, sparse=False):
    """
    Return output for variable `var` as a pandas.Series (1d),
    pandas.Dataframe (2d), or xarray.DataArray (3d and higher).

    Parameters
    ----------
    var : variable name as string, e.g. 'resource'
    dims : list, optional
        indices as strings, e.g. ('loc_techs', 'timesteps');
        if not given, they are auto-detected
    sparse : bool, optional; default = False
        If extracting Pyomo Param data, the output sparse array includes inputs
        the user left as NaN replaced with the default value for that Param.
    """
    try:
        var_container = getattr(backend_model, var)
    except AttributeError:
        raise exceptions.BackendError('Variable {} inexistent.'.format(var))

    if not dims:
        if var + '_index' == var_container.index_set().name:
            dims = [i.name for i in var_container.index_set().set_tuple]
        else:
            dims = [var_container.index_set().name]

    if sparse:
        result = pd.DataFrame.from_dict(var_container.extract_values_sparse(), orient='index')
    else:
        result = pd.DataFrame.from_dict(var_container.extract_values(), orient='index')

    if result.empty:
        raise exceptions.BackendError('Variable {} has no data.'.format(var))

    result = result[0]  # Get the only column in the dataframe

    if len(dims) > 1:
        result.index = pd.MultiIndex.from_tuples(result.index, names=dims)

    if len(result.index.names) == 1:
        result = result.sort_index()
        result.index.name = dims[0]
    elif len(result.index.names) == 2:
        # if len(dims) is 2, we already have a well-formed DataFrame
        result = result.unstack(level=0)
        result = result.sort_index()
    else:  # len(dims) >= 3
        result = xr.DataArray.from_series(result)

    return result


@memoize
def loc_tech_is_in(backend_model, loc_tech, model_set):
    """
    Check if set exists and if loc_tech is in the set

    Parameters
    ----------
    loc_tech : string
    model_set : string
    """

    if hasattr(backend_model, model_set) and loc_tech in getattr(backend_model, model_set):
        return True
    else:
        return False




"""
Copyright (C) 2013-2018 Calliope contributors listed in AUTHORS.
Licensed under the Apache 2.0 License (see LICENSE file).

logger.py
~~~~~~~~~~

Create the Calliope logger object and apply other logging tools/functionality

"""

import logging
import sys
import datetime

SOLVER = 19
logging.addLevelName(SOLVER, 'SOLVER')
logger = logging.getLogger('calliope')
logger.propagate = False

if logger.hasHandlers():
    for handler in logger.handlers:
        logger.removeHandler(handler)

formatter = logging.Formatter(
    '[%(asctime)s] %(levelname)s: %(message)s',
    datefmt="%Y-%m-%d %H:%M:%S")
console = logging.StreamHandler(stream=sys.stdout)
console.setFormatter(formatter)
logger.addHandler(console)


def set_log_level(level):
    """
    Set the minimum logging verbosity in a Python console. Higher verbosity levels
    will include their output and all those of following levels.
    Level options (in descending order of verbosity):

    * 'DEBUG'
    * 'SOLVER' -> Calliope custom level, assigned value of 19,
                  returns solver (e.g. GLPK) stream
    * 'INFO'
    * 'WARNING' -> default level
    * 'ERROR'
    * 'CRITICAL'
    """

    if level == 'DEBUG':
        logger.setLevel(logging.DEBUG)

    elif level == 'SOLVER':
        logger.setLevel(SOLVER)

    else:
        logger.setLevel(getattr(logging, level))


def log_time(timings, identifier, comment=None, level='info', time_since_start=False):
    if comment is None:
        comment = identifier

    timings[identifier] = now = datetime.datetime.now()

    if time_since_start:
        time_diff = now - timings['model_creation']
        comment += '. Time since start: {}'.format(time_diff)

    getattr(logger, level)(comment)


class LogWriter:
    def __init__(self, level, strip=False):
        self.level = level
        self.strip = strip

    def write(self, message):
        if message != '\n':
            if self.strip:
                message = message.strip()
            if self.level == 'solver':
                logger.log(SOLVER, message)
            else:
                getattr(logger, self.level)(message)

    def flush(self):
        pass



"""
Copyright (C) 2013-2018 Calliope contributors listed in AUTHORS.
Licensed under the Apache 2.0 License (see LICENSE file).

"""

from calliope import exceptions
import xarray as xr
import pandas as pd


def get_loc_techs(loc_techs, tech=None, loc=None):
    """
    Get a list of loc_techs associated with the given technology and/or location.
    If multiple of both loc and tech are given, the function will return any
    combination of members of loc and tech lists found in loc_techs.

    Parameters
    ----------
    loc_techs : list
        set of loc_techs to search for the relevant tech and/or loc
    tech : string or list of strings, default None
        technology/technologies to search for in the set of location:technology
    loc : string or list of strings, default None
        location(s) to search for in the set of location:technology

    Returns
    -------
    relevant_loc_techs : list of strings

    """
    # If both are strings, there is only one loc:tech possibility to look for
    if (isinstance(tech, str) and isinstance(loc, str)
        and '::'.join([loc, tech]) in loc_techs):
        relevant_loc_techs = ['::'.join([loc, tech])]

    tech = [tech] if tech is not None and isinstance(tech, str) else tech
    loc = [loc] if loc is not None and isinstance(loc, str) else loc

    if tech and not loc:
        relevant_loc_techs = [i for i in loc_techs if i.split('::')[1] in tech]
    elif loc and not tech:
        relevant_loc_techs = [i for i in loc_techs if i.split('::')[0] in loc]
    elif loc and tech:
        loc_techs_set = set(tuple(i.split('::')) for i in loc_techs)
        possible_loc_techs = set((l, t) for l in loc for t in tech)
        relevant_loc_techs = [
            '::'.join(i) for i in possible_loc_techs.intersection(loc_techs_set)
        ]
    else:
        relevant_loc_techs = [None]

    return relevant_loc_techs


def reorganise_xarray_dimensions(data):
    """
    Reorganise Dataset or DataArray dimensions to be alphabetical *except*
    `timesteps`, which must always come last in any DataArray's dimensions
    """

    if not (isinstance(data, xr.Dataset) or isinstance(data, xr.DataArray)):
        raise TypeError('Must provide either xarray Dataset or DataArray to be reorganised')

    steps = [i for i in ['datesteps', 'timesteps'] if i in data.dims]

    if isinstance(data, xr.Dataset):
        new_dims = (
            sorted(list(set(data.dims.keys()) - set(steps)))
        ) + steps
    elif isinstance(data, xr.DataArray):
        new_dims = (
            sorted(list(set(data.dims) - set(steps)))
        ) + steps

    updated_data = data.transpose(*new_dims).reindex({k: data[k] for k in new_dims})

    return updated_data


def split_loc_techs(data_var, return_as='DataArray'):
    """
    Get a DataArray with locations technologies, and possibly carriers
    split into separate coordinates.

    Parameters
    ----------
    data_var : xarray DataArray
        Variable from Calliope model_data, to split loc_techs dimension
    return_as : string
        'DataArray' to return xarray DataArray, 'MultiIndex DataArray' to return
        xarray DataArray with loc_techs as a MultiIndex,
        or 'Series' to return pandas Series with dimensions as a MultiIndex

    Returns
    -------
    updated_data_var : xarray DataArray of pandas Series
    """

    # Separately find the loc_techs(_carriers) dimension and all other dimensions
    loc_tech_dim = [i for i in data_var.dims if 'loc_tech' in i]
    if not loc_tech_dim:
        loc_tech_dim = [i for i in data_var.dims if 'loc_carrier' in i]

    if not loc_tech_dim:
        if return_as == 'Series':
            return data_var.to_series()
        elif return_as in ['DataArray', 'MultiIndex DataArray']:
            return data_var
        else:
            raise ValueError('`return_as` must be `DataArray`, `Series`, or '
                             '`MultiIndex DataArray`, but `{}` given'.format(return_as))

    elif len(loc_tech_dim) > 1:
        e = exceptions.ModelError
        raise e("Cannot split loc_techs or loc_tech_carriers dimension "
                "for DataArray {}".format(data_var.name))

    loc_tech_dim = loc_tech_dim[0]
    # xr.Datarray -> pd.Series allows for string operations
    data_var_idx = data_var[loc_tech_dim].to_index()
    index_list = data_var_idx.str.split('::').tolist()

    # carrier_prod, carrier_con, and carrier_export will return an index_list
    # of size 3, all others will be an index list of size 2
    possible_names = ['loc', 'tech', 'carrier']
    names = [i + 's' for i in possible_names if i in loc_tech_dim]

    data_var_midx = pd.MultiIndex.from_tuples(index_list, names=names)

    # Replace the Datarray loc_tech_dim with this new MultiIndex
    updated_data_var = data_var.copy()
    updated_data_var.coords[loc_tech_dim] = data_var_midx

    if return_as == 'MultiIndex DataArray':
        return updated_data_var

    elif return_as == "Series":
        return reorganise_xarray_dimensions(updated_data_var.unstack()).to_series()

    elif return_as == "DataArray":
        return reorganise_xarray_dimensions(updated_data_var.unstack())

    else:
        raise ValueError('`return_as` must be `DataArray`, `Series`, or '
                         '`MultiIndex DataArray`, but `{}` given'.format(return_as))



import os

import pytest
import tempfile
import pandas as pd

import calliope
from calliope.test.common.util import check_error_or_warning

class TestModel:
    @pytest.fixture(scope="module")
    def national_scale_example(self):
        model = calliope.examples.national_scale(
            override_dict={'model.subset_time': '2005-01-01'}
        )
        model.run()
        return model

    def test_save_commented_model_yaml(self, national_scale_example):
        model = national_scale_example

        with tempfile.TemporaryDirectory() as tempdir:
            out_path = os.path.join(tempdir, 'model_debug.yaml')
            model.save_commented_model_yaml(out_path)
            assert os.path.isfile(out_path)

    def test_info(self, national_scale_example):
        model = national_scale_example

        model.info()

    def test_info_minimal_model(self):
        this_path = os.path.dirname(__file__)
        model_location = os.path.join(this_path, 'common', 'test_model', 'model_minimal.yaml')
        model = calliope.Model(model_location)

        model.info()

    def test_get_formatted_array_index(self, national_scale_example):
        array = national_scale_example.get_formatted_array('resource', index_format='index')

        assert array.dims == ('locs', 'techs', 'timesteps')

    def test_get_formatted_array_multiindex(self, national_scale_example):
        array = national_scale_example.get_formatted_array('resource', index_format='multiindex')

        assert array.dims == ('loc_techs_finite_resource', 'timesteps')
        assert isinstance(array.loc_techs_finite_resource.to_index(), pd.MultiIndex)

    def test_get_formatted_array_unknown_format(self, national_scale_example):
        with pytest.raises(ValueError) as excinfo:
            national_scale_example.get_formatted_array('resource', index_format='foo')

        assert check_error_or_warning(
            excinfo, "Argument 'index_format' must be one of 'index' or 'multiindex'"
        )

    def test_get_formatted_array_unknown_var(self, national_scale_example):
        with pytest.raises(KeyError) as excinfo:
            national_scale_example.get_formatted_array('foo')

        assert check_error_or_warning(
            excinfo, "Variable foo not in Model data"
        )



"""
Copyright (C) 2013-2018 Calliope contributors listed in AUTHORS.
Licensed under the Apache 2.0 License (see LICENSE file).

constraint_sets.py
~~~~~~~~~~~~~~~~~~

loc_techs, loc_carriers, and loc_tech_carriers subsets used per constraint, to
reduce constraint complexity

"""

from itertools import product

import numpy as np

from calliope.core.preprocess.util import constraint_exists, concat_iterable


def generate_constraint_sets(model_run):
    """
    Generate loc-tech sets for a given pre-processed ``model_run``

    Parameters
    ----------
    model_run : AttrDict
    """

    sets = model_run.sets
    ## From here on, everything is a `key=value` pair within a dictionary

    constraint_sets = dict()
    # energy_balance.py
    constraint_sets['loc_carriers_system_balance_constraint'] = sets.loc_carriers
    constraint_sets['loc_techs_balance_supply_constraint'] = sets.loc_techs_finite_resource_supply
    constraint_sets['loc_techs_balance_demand_constraint'] = sets.loc_techs_finite_resource_demand
    constraint_sets['loc_techs_resource_availability_supply_plus_constraint'] = sets.loc_techs_finite_resource_supply_plus
    constraint_sets['loc_techs_balance_transmission_constraint'] = sets.loc_techs_transmission
    constraint_sets['loc_techs_balance_supply_plus_constraint'] = sets.loc_techs_supply_plus
    constraint_sets['loc_techs_balance_storage_constraint'] = sets.loc_techs_storage
    if model_run.run.cyclic_storage is True:
        constraint_sets['loc_techs_storage_initial_constraint'] = [
            i for i in sets.loc_techs_store
            if constraint_exists(model_run, i, 'constraints.storage_initial') is not None
        ]
    constraint_sets['carriers_reserve_margin_constraint'] = [
        i for i in sets.carriers
        if i in model_run.model.get_key('reserve_margin', {}).keys()
    ]
    # clustering-specific balance constraints
    if (model_run.model.get_key('time.function', None) == 'apply_clustering' and
            model_run.model.get_key('time.function_options.storage_inter_cluster', True)):
        set_name = 'loc_techs_balance_storage_inter_cluster_constraint'
        constraint_sets[set_name] = sets.loc_techs_store

    # costs.py
    constraint_sets['loc_techs_cost_constraint'] = sets.loc_techs_cost
    constraint_sets['loc_techs_cost_investment_constraint'] = sets.loc_techs_investment_cost
    constraint_sets['loc_techs_cost_var_constraint'] = [
        i for i in sets.loc_techs_om_cost
        if i not in sets.loc_techs_conversion_plus + sets.loc_techs_conversion
    ]

    # export.py
    constraint_sets['loc_carriers_update_system_balance_constraint'] = [
        i for i in sets.loc_carriers if sets.loc_techs_export
        and any(['{0}::{2}'.format(*j.split('::')) == i
                for j in sets.loc_tech_carriers_export])
    ]
    constraint_sets['loc_tech_carriers_export_balance_constraint'] = (
        sets.loc_tech_carriers_export
    )
    constraint_sets['loc_techs_update_costs_var_constraint'] = [
        i for i in sets.loc_techs_om_cost if i in sets.loc_techs_export
    ]

    constraint_sets['loc_tech_carriers_export_max_constraint'] = [
        i for i in sets.loc_tech_carriers_export
        if constraint_exists(
            model_run, i.rsplit('::', 1)[0], 'constraints.export_cap'
        ) is not None
    ]

    # capacity.py
    constraint_sets['loc_techs_storage_capacity_constraint'] = [
        i for i in sets.loc_techs_store if i not in sets.loc_techs_milp
    ]
    constraint_sets['loc_techs_energy_capacity_storage_constraint_old'] = [
        i for i in sets.loc_techs_store
        if constraint_exists(model_run, i, 'constraints.charge_rate')
    ]
    constraint_sets['loc_techs_energy_capacity_storage_equals_constraint'] = [
        i for i in sets.loc_techs_store
        if constraint_exists(model_run, i, 'constraints.energy_cap_per_storage_cap_equals')
    ]
    constraint_sets['loc_techs_energy_capacity_storage_min_constraint'] = [
        i for i in sets.loc_techs_store
        if constraint_exists(model_run, i, 'constraints.energy_cap_per_storage_cap_min')
        and not constraint_exists(model_run, i, 'constraints.energy_cap_per_storage_cap_equals')
    ]
    constraint_sets['loc_techs_energy_capacity_storage_max_constraint'] = [
        i for i in sets.loc_techs_store
        if constraint_exists(model_run, i, 'constraints.energy_cap_per_storage_cap_max')
        and not constraint_exists(model_run, i, 'constraints.energy_cap_per_storage_cap_equals')
    ]
    constraint_sets['loc_techs_resource_capacity_constraint'] = [
        i for i in sets.loc_techs_finite_resource_supply_plus
        if any([
            constraint_exists(model_run, i, 'constraints.resource_cap_equals'),
            constraint_exists(model_run, i, 'constraints.resource_cap_max'),
            constraint_exists(model_run, i, 'constraints.resource_cap_min')
        ])
    ]
    constraint_sets['loc_techs_resource_capacity_equals_energy_capacity_constraint'] = [
        i for i in sets.loc_techs_finite_resource_supply_plus
        if constraint_exists(model_run, i, 'constraints.resource_cap_equals_energy_cap')
    ]
    constraint_sets['loc_techs_resource_area_constraint'] = sets.loc_techs_area
    constraint_sets['loc_techs_resource_area_per_energy_capacity_constraint'] = [
        i for i in sets.loc_techs_area
        if constraint_exists(model_run, i, 'constraints.resource_area_per_energy_cap')
        is not None
    ]
    constraint_sets['locs_resource_area_capacity_per_loc_constraint'] = [
        i for i in sets.locs
        if model_run.locations[i].get_key('available_area', None) is not None
        and sets.loc_techs_area
    ]
    constraint_sets['loc_techs_energy_capacity_constraint'] = [
        i for i in sets.loc_techs
        if i not in sets.loc_techs_milp + sets.loc_techs_purchase
    ]
    constraint_sets['techs_energy_capacity_systemwide_constraint'] = [
        i for i in sets.techs
        if model_run.get_key('techs.{}.constraints.energy_cap_max_systemwide'.format(i), None)
        or model_run.get_key('techs.{}.constraints.energy_cap_equals_systemwide'.format(i), None)
    ]

    # dispatch.py
    constraint_sets['loc_tech_carriers_carrier_production_max_constraint'] = [
        i for i in sets.loc_tech_carriers_prod
        if i not in sets.loc_tech_carriers_conversion_plus
        and i.rsplit('::', 1)[0] not in sets.loc_techs_milp
    ]
    constraint_sets['loc_tech_carriers_carrier_production_min_constraint'] = [
        i for i in sets.loc_tech_carriers_prod
        if i not in sets.loc_tech_carriers_conversion_plus
        and constraint_exists(model_run, i.rsplit('::', 1)[0], 'constraints.energy_cap_min_use')
        and i.rsplit('::', 1)[0] not in sets.loc_techs_milp
    ]
    constraint_sets['loc_tech_carriers_carrier_consumption_max_constraint'] = [
        i for i in sets.loc_tech_carriers_con
        if i.rsplit('::', 1)[0] in sets.loc_techs_demand +
            sets.loc_techs_storage + sets.loc_techs_transmission
        and i.rsplit('::', 1)[0] not in sets.loc_techs_milp
    ]
    constraint_sets['loc_techs_resource_max_constraint'] = sets.loc_techs_supply_plus
    constraint_sets['loc_tech_carriers_ramping_constraint'] = [
        i for i in sets.loc_tech_carriers_prod
        if i.rsplit('::', 1)[0] in sets.loc_techs_ramping
    ]
    # clustering-specific dispatch constraints
    if (model_run.model.get_key('time.function', None) == 'apply_clustering' and
            model_run.model.get_key('time.function_options.storage_inter_cluster', True)):
        constraint_sets['loc_techs_storage_intra_max_constraint'] = sets.loc_techs_store
        constraint_sets['loc_techs_storage_intra_min_constraint'] = sets.loc_techs_store
        constraint_sets['loc_techs_storage_inter_max_constraint'] = sets.loc_techs_store
        constraint_sets['loc_techs_storage_inter_min_constraint'] = sets.loc_techs_store
    else:
        constraint_sets['loc_techs_storage_max_constraint'] = sets.loc_techs_store

    # milp.py
    constraint_sets['loc_techs_unit_commitment_milp_constraint'] = sets.loc_techs_milp
    constraint_sets['loc_techs_unit_capacity_milp_constraint'] = sets.loc_techs_milp
    constraint_sets['loc_tech_carriers_carrier_production_max_milp_constraint'] = [
        i for i in sets.loc_tech_carriers_prod
        if i not in sets.loc_tech_carriers_conversion_plus
        and i.rsplit('::', 1)[0] in sets.loc_techs_milp
    ]
    constraint_sets['loc_techs_carrier_production_max_conversion_plus_milp_constraint'] = [
        i for i in sets.loc_techs_conversion_plus
        if i in sets.loc_techs_milp
    ]
    constraint_sets['loc_tech_carriers_carrier_production_min_milp_constraint'] = [
        i for i in sets.loc_tech_carriers_prod
        if i not in sets.loc_tech_carriers_conversion_plus
        and constraint_exists(model_run, i.rsplit('::', 1)[0], 'constraints.energy_cap_min_use')
        and i.rsplit('::', 1)[0] in sets.loc_techs_milp
    ]
    constraint_sets['loc_techs_carrier_production_min_conversion_plus_milp_constraint'] = [
        i for i in sets.loc_techs_conversion_plus
        if constraint_exists(model_run, i, 'constraints.energy_cap_min_use')
        and i in sets.loc_techs_milp
    ]
    constraint_sets['loc_tech_carriers_carrier_consumption_max_milp_constraint'] = [
        i for i in sets.loc_tech_carriers_con
        if i.rsplit('::', 1)[0] in sets.loc_techs_demand +
            sets.loc_techs_storage + sets.loc_techs_transmission
        and i.rsplit('::', 1)[0] in sets.loc_techs_milp
    ]
    constraint_sets['loc_techs_energy_capacity_units_milp_constraint'] = [
        i for i in sets.loc_techs_milp
        if constraint_exists(model_run, i, 'constraints.energy_cap_per_unit')
        is not None
    ]
    constraint_sets['loc_techs_storage_capacity_units_milp_constraint'] = [
        i for i in sets.loc_techs_milp if i in sets.loc_techs_store
    ]
    constraint_sets['loc_techs_energy_capacity_max_purchase_milp_constraint'] = [
        i for i in sets.loc_techs_purchase
        if (constraint_exists(model_run, i, 'constraints.energy_cap_equals') is not None
            or (constraint_exists(model_run, i, 'constraints.energy_cap_max') is not None
                and constraint_exists(model_run, i, 'constraints.energy_cap_max') != np.inf))
    ]
    constraint_sets['loc_techs_energy_capacity_min_purchase_milp_constraint'] = [
        i for i in sets.loc_techs_purchase
        if (not constraint_exists(model_run, i, 'constraints.energy_cap_equals')
            and constraint_exists(model_run, i, 'constraints.energy_cap_min'))
    ]
    constraint_sets['loc_techs_storage_capacity_max_purchase_milp_constraint'] = [
        i for i in set(sets.loc_techs_purchase).intersection(sets.loc_techs_store)
        if (constraint_exists(model_run, i, 'constraints.storage_cap_equals') is not None
            or (constraint_exists(model_run, i, 'constraints.storage_cap_max') is not None
                and constraint_exists(model_run, i, 'constraints.storage_cap_max') != np.inf))
    ]
    constraint_sets['loc_techs_storage_capacity_min_purchase_milp_constraint'] = [
        i for i in set(sets.loc_techs_purchase).intersection(sets.loc_techs_store)
        if (not constraint_exists(model_run, i, 'constraints.storage_cap_equals')
            and constraint_exists(model_run, i, 'constraints.storage_cap_min'))
    ]
    constraint_sets['loc_techs_update_costs_investment_units_milp_constraint'] = [
        i for i in sets.loc_techs_milp
        if i in sets.loc_techs_investment_cost and
        any(constraint_exists(model_run, i, 'costs.{}.purchase'.format(j))
            for j in model_run.sets.costs)
    ]
    # loc_techs_purchase technologies only exist because they have defined a purchase cost
    constraint_sets['loc_techs_update_costs_investment_purchase_milp_constraint'] = sets.loc_techs_purchase

    constraint_sets['techs_unit_capacity_systemwide_milp_constraint'] = [
        i for i in sets.techs
        if model_run.get_key('techs.{}.constraints.units_max_systemwide'.format(i), None)
        or model_run.get_key('techs.{}.constraints.units_equals_systemwide'.format(i), None)
    ]
    constraint_sets['loc_techs_asynchronous_prod_con_milp_constraint'] = sets.loc_techs_asynchronous_prod_con

    # conversion.py
    constraint_sets['loc_techs_balance_conversion_constraint'] = sets.loc_techs_conversion
    constraint_sets['loc_techs_cost_var_conversion_constraint'] = sets.loc_techs_om_cost_conversion

    # conversion_plus.py
    constraint_sets['loc_techs_balance_conversion_plus_primary_constraint'] = sets.loc_techs_conversion_plus
    constraint_sets['loc_techs_carrier_production_max_conversion_plus_constraint'] = [
        i for i in sets.loc_techs_conversion_plus
        if i not in sets.loc_techs_milp
    ]
    constraint_sets['loc_techs_carrier_production_min_conversion_plus_constraint'] = [
        i for i in sets.loc_techs_conversion_plus
        if constraint_exists(model_run, i, 'constraints.energy_cap_min_use')
        and i not in sets.loc_techs_milp
    ]
    constraint_sets['loc_techs_cost_var_conversion_plus_constraint'] = sets.loc_techs_om_cost_conversion_plus
    constraint_sets['loc_techs_balance_conversion_plus_in_2_constraint'] = sets.loc_techs_in_2
    constraint_sets['loc_techs_balance_conversion_plus_in_3_constraint'] = sets.loc_techs_in_3
    constraint_sets['loc_techs_balance_conversion_plus_out_2_constraint'] = sets.loc_techs_out_2
    constraint_sets['loc_techs_balance_conversion_plus_out_3_constraint'] = sets.loc_techs_out_3

    # network.py
    constraint_sets['loc_techs_symmetric_transmission_constraint'] = sets.loc_techs_transmission

    # policy.py
    constraint_sets['techlists_group_share_energy_cap_min_constraint'] = [
        i for i in sets.techlists
        if 'energy_cap_min' in model_run.model.get_key('group_share.{}'.format(i), {}).keys()
    ]
    constraint_sets['techlists_group_share_energy_cap_max_constraint'] = [
        i for i in sets.techlists
        if 'energy_cap_max' in model_run.model.get_key('group_share.{}'.format(i), {}).keys()
    ]
    constraint_sets['techlists_group_share_energy_cap_equals_constraint'] = [
        i for i in sets.techlists
        if 'energy_cap_equals' in model_run.model.get_key('group_share.{}'.format(i), {}).keys()
    ]
    constraint_sets['techlists_carrier_group_share_carrier_prod_min_constraint'] = [
        i + '::' + carrier
        for i in sets.techlists
        if 'carrier_prod_min' in model_run.model.get_key('group_share.{}'.format(i), {}).keys()
        for carrier in sets.carriers
        if carrier in model_run.model.get_key('group_share.{}.carrier_prod_min'.format(i), {}).keys()
    ]
    constraint_sets['techlists_carrier_group_share_carrier_prod_max_constraint'] = [
        i + '::' + carrier
        for i in sets.techlists
        if 'carrier_prod_max' in model_run.model.get_key('group_share.{}'.format(i), {}).keys()
        for carrier in sets.carriers
        if carrier in model_run.model.get_key('group_share.{}.carrier_prod_max'.format(i), {}).keys()
    ]
    constraint_sets['techlists_carrier_group_share_carrier_prod_equals_constraint'] = [
        i + '::' + carrier
        for i in sets.techlists
        if 'carrier_prod_equals' in model_run.model.get_key('group_share.{}'.format(i), {}).keys()
        for carrier in sets.carriers
        if carrier in model_run.model.get_key('group_share.{}.carrier_prod_equals'.format(i), {}).keys()
    ]

    # group.py
    group_constraints = {
        name: data for name, data in model_run['group_constraints'].items()
        if data.get("exists", True)
    }
    constraint_sets['constraint_groups'] = list(group_constraints.keys())

    for k, v in group_constraints.items():
        # For now, transmission techs are not supported in group constraints
        techs = v.get('techs', sets['techs_non_transmission'])
        locs = v.get('locs', sets['locs'])

        # All possible loc_techs for this constraint
        loc_techs_all = list(set(concat_iterable(
            [(l, t) for l, t in product(locs, techs)],
            ['::']
        )))

        # Some loc_techs may not actually exist in the actual model,
        # so we must filter with actually exising loc_techs
        loc_techs = [i for i in loc_techs_all if i in sets.loc_techs]

        constraint_sets['group_constraint_loc_techs_{}'.format(k)] = loc_techs

    return constraint_sets
